import os

from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QComboBox, QFileDialog, QStyle


class FolderComboBox(QComboBox):
    currentFolderModified = pyqtSignal(str)

    def __init__(self, parent=None):
        super().__init__(parent)
        self._currentFolder = None
        # TODO: could be done in showEvent only if currentFolder is
        # None/untouched
        self._updateContents()

        self.currentIndexChanged[int].connect(self._updateCurrentFolder)

    def currentFolder(self):
        return self._currentFolder

    def setCurrentFolder(self, path):
        if path == self._currentFolder:
            return
        if path is not None:
            assert os.path.isdir(path)
        self._currentFolder = path
        self._updateContents()
        self.currentFolderModified.emit(self._currentFolder)

    def _updateContents(self):
        self.blockSignals(True)
        self.clear()
        path = self._currentFolder
        if path is not None:
            dirIcon = self.style().standardIcon(QStyle.SP_DirIcon)
            self.addItem(dirIcon, os.path.basename(path))
        else:
            self.addItem("<None>")
        self.insertSeparator(1)
        self.addItem(self.tr("Browse…"))
        self.setCurrentIndex(0)
        self.blockSignals(False)

    def _updateCurrentFolder(self, index):
        if index < self.count() - 1:
            return
        # TODO: use app.activeWindow()?
        path = QFileDialog.getExistingDirectory(
            self,
            self.tr("Choose Directory"),
            self._currentFolder,
            QFileDialog.ShowDirsOnly,
        )
        if path:
            self.setCurrentFolder(path)
        else:
            self.setCurrentIndex(0)



"""
The *glyphSequenceEdit* submodule
-----------------------------

The *glyphSequenceEdit* submodule provides a QLineEdit_ widget that besides its
text string can return a list of Glyph_ from the font specified in its
constructor.

.. _Glyph: http://ts-defcon.readthedocs.org/en/ufo3/objects/glyph.html
.. _QLineEdit: http://doc.qt.io/qt-5/qlineedit.html
"""

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QComboBox, QLineEdit

from defconQt.tools.textSplitter import splitText


def _glyphs(self):
    text = self.text()
    glyphNames = self.splitTextFunction(text, self._font.unicodeData)
    glyphs = [
        self._font[glyphName] for glyphName in glyphNames if glyphName in self._font
    ]
    return glyphs


class GlyphSequenceComboBox(QComboBox):
    splitTextFunction = staticmethod(splitText)

    def __init__(self, font, parent=None):
        super().__init__(parent)
        # setEditable(True) must be called before self.completer()
        # otherwise it will return None
        self.setEditable(True)
        completer = self.completer()
        completer.setCaseSensitivity(Qt.CaseSensitive)
        self.setCompleter(completer)
        self._font = font

    glyphs = _glyphs

    def text(self):
        return self.currentText()

    def setText(self, text):
        self.setEditText(text)


class GlyphSequenceEdit(QLineEdit):
    splitTextFunction = staticmethod(splitText)

    def __init__(self, font, parent=None):
        super().__init__(parent)
        self._font = font

    glyphs = _glyphs



import os
import sys

from PyQt5.QtCore import QCommandLineParser, QLibraryInfo, QLocale, Qt, QTranslator
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QApplication

from defconQt import representationFactories as baseRepresentationFactories
from trufont import __version__, representationFactories
from trufont.objects import settings
from trufont.objects.application import Application
from trufont.objects.extension import TExtension
from trufont.resources import icons_db  # noqa
from trufont.tools import errorReports, platformSpecific
from trufont.windows.outputWindow import OutputWindow


def main():
    global app

    # register representation factories
    baseRepresentationFactories.registerAllFactories()
    representationFactories.registerAllFactories()
    if hasattr(Qt, "AA_EnableHighDpiScaling"):
        QApplication.setAttribute(Qt.AA_EnableHighDpiScaling)
    platformSpecific.setAppName()
    # initialize the app
    app = Application(sys.argv)
    app.setOrganizationName("TruFont")
    app.setOrganizationDomain("trufont.github.io")
    app.setApplicationName("TruFont")
    app.setApplicationVersion(__version__)
    app.setWindowIcon(QIcon(":app.png"))
    app.setAttribute(Qt.AA_UseHighDpiPixmaps, True)
    appFont = platformSpecific.UIFontOverride()
    if appFont is not None:
        app.setFont(appFont)
    app.setStyleSheet(platformSpecific.appStyleSheet())

    # Install stream redirection
    app.outputWindow = OutputWindow()
    # Exception handling
    sys.excepthook = errorReports.exceptionCallback

    # Qt's translation for itself. May not be installed.
    qtTranslator = QTranslator()
    qtTranslator.load(
        "qt_" + QLocale.system().name(),
        QLibraryInfo.location(QLibraryInfo.TranslationsPath),
    )
    app.installTranslator(qtTranslator)

    appTranslator = QTranslator()
    appTranslator.load(
        "trufont_" + QLocale.system().name(),
        os.path.dirname(os.path.realpath(__file__)) + "/resources",
    )
    app.installTranslator(appTranslator)

    # parse options and open fonts
    parser = QCommandLineParser()
    parser.setApplicationDescription(
        QApplication.translate("Command-line parser", "The TruFont font editor.")
    )
    parser.addHelpOption()
    parser.addVersionOption()
    parser.addPositionalArgument(
        QApplication.translate("Command-line parser", "files"),
        QApplication.translate("Command-line parser", "The UFO files to open."),
    )
    parser.process(app)
    # load menu
    if platformSpecific.useGlobalMenuBar():
        app.fetchMenuBar()
        app.setQuitOnLastWindowClosed(False)
    # bootstrap extensions
    folder = app.getExtensionsDirectory()
    for file in os.listdir(folder):
        if not file.rstrip("\\/ ").endswith(".tfExt"):
            continue
        path = os.path.join(folder, file)
        try:
            extension = TExtension(path)
            if extension.launchAtStartup:
                extension.run()
        except Exception as e:
            msg = QApplication.translate(
                "Extensions", f"The extension at {path} could not be run."
            )
            errorReports.showWarningException(e, msg)
            continue
        app.registerExtension(extension)
    # process files
    args = parser.positionalArguments()
    if not args:
        # maybe load recent file
        loadRecentFile = settings.loadRecentFile()
        if loadRecentFile:
            recentFiles = settings.recentFiles()
            if len(recentFiles) and os.path.exists(recentFiles[0]):
                app.openFile(recentFiles[0])
    else:
        for fontPath in args:
            app.openFile(fontPath)
    # if we did not open a font, spawn new font or go headless
    if not app.allFonts():
        if platformSpecific.shouldSpawnDocument():
            app.newFile()
        else:
            # HACK: on OSX we may want to trigger native QMenuBar display
            # without opening any window. Since Qt infers new menu bar on
            # focus change, fire the signal.
            app.focusWindowChanged.emit(None)
    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



import itertools
from collections import OrderedDict

from PyQt5.QtCore import QLineF, QPointF, Qt
from PyQt5.QtGui import QColor, QPainterPath
from PyQt5.QtWidgets import QApplication

from trufont.drawingTools.baseTool import BaseTool
from trufont.tools import bezierMath, drawing

_path = QPainterPath()
_path.moveTo(4.1, 19.94)
_path.lineTo(19.94, 4.1)
_path.lineTo(23.9, 8.06)
_path.lineTo(22.25, 9.71)
_path.lineTo(20.92, 8.4)
_path.lineTo(19.94, 9.38)
_path.lineTo(21.26, 10.7)
_path.lineTo(19.94, 12.02)
_path.lineTo(17.63, 9.71)
_path.lineTo(16.64, 10.7)
_path.lineTo(18.95, 13.01)
_path.lineTo(17.63, 14.33)
_path.lineTo(16.31, 13.01)
_path.lineTo(15.32, 14)
_path.lineTo(16.64, 15.32)
_path.lineTo(15.32, 16.64)
_path.lineTo(13.01, 14.33)
_path.lineTo(12.02, 15.32)
_path.lineTo(14.33, 17.63)
_path.lineTo(13.01, 18.95)
_path.lineTo(11.69, 17.63)
_path.lineTo(10.7, 18.62)
_path.lineTo(12.02, 19.94)
_path.lineTo(10.7, 21.26)
_path.lineTo(8.4, 18.95)
_path.lineTo(7.4, 19.94)
_path.lineTo(9.71, 22.25)
_path.lineTo(8.06, 23.9)
_path.closeSubpath()


class RulerTool(BaseTool):
    icon = _path
    name = QApplication.translate("RulerTool", "Ruler")
    shortcut = "L"

    def __init__(self, parent=None):
        super().__init__(parent)
        self._cachedIntersections = None
        self._rulerObject = None
        self._rulerPts = None

    def toolActivated(self):
        self.parent().update()

    def toolDisabled(self):
        self._cachedIntersections = None
        self._rulerObject = None
        self._rulerPts = None
        self.parent().update()

    def drawingAttribute(self, attr, flags):
        if flags.isActiveLayer and attr == "showGlyphPointCoordinates":
            return True
        return super().drawingAttribute(attr, flags)

    # custom methods

    def _appendIntersection(self, pt):
        x, y, _ = pt
        line = QLineF(self._rulerObject[0])
        line.setP2(QPointF(x, y))
        self._rulerPts[line.length()] = (x, y)

    def _findIntersections(self):
        self._cachedIntersections = OrderedDict()
        self._rulerPts = dict()
        line = self._rulerObject[0]

        # the default layer stores the width
        defaultGlyph = self._glyph.layerSet.defaultLayer.get(self._glyph.name)
        if defaultGlyph is not None:
            width = defaultGlyph.width
            info = defaultGlyph.font.info
            ascender = info.ascender or 750
            descender = info.descender or -250
            for x in (0, width):
                pt = bezierMath.lineIntersection(
                    line.x1(),
                    line.y1(),
                    line.x2(),
                    line.y2(),
                    x,
                    descender,
                    x,
                    ascender,
                )
                if pt is not None:
                    self._appendIntersection(pt)

        for contour in self._glyph:
            segments = contour.segments
            for index, seg in enumerate(segments):
                if seg[-1].segmentType == "move":
                    continue
                prev = segments[index - 1][-1]
                if seg[-1].segmentType == "curve":
                    i = bezierMath.curveIntersections(
                        line.x1(),
                        line.y1(),
                        line.x2(),
                        line.y2(),
                        prev,
                        seg[0],
                        seg[1],
                        seg[2],
                    )
                    for pt in i:
                        self._appendIntersection(pt)
                elif len(seg) == 1:
                    pt = bezierMath.lineIntersection(
                        line.x1(),
                        line.y1(),
                        line.x2(),
                        line.y2(),
                        prev.x,
                        prev.y,
                        seg[0].x,
                        seg[0].y,
                    )
                    if pt is not None:
                        self._appendIntersection(pt)

    # events

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            pos = self.magnetPos(event.localPos())
            line = QLineF(pos, pos)
            self._rulerObject = (line, "0.0º")
            self._rulerPts = dict()
            self.parent().update()
        else:
            super().mousePressEvent(event)

    def mouseMoveEvent(self, event):
        if event.buttons() & Qt.LeftButton:
            pos = self.magnetPos(event.localPos())
            if self._rulerObject is None:
                self._rulerObject = (QLineF(pos, pos), "0.0º")
                return
            line, _ = self._rulerObject
            # magnet done before clamping to axis
            if event.modifiers() & Qt.ShiftModifier:
                pos = self.clampToOrigin(pos, line.p1())
            line.setP2(pos)
            # angle() doesnt go by trigonometric direction. Weird.
            angle = line.angle()
            angle = 360 * (angle >= 180) - angle
            a = "{}º".format(round(angle, 1))
            self._rulerObject = (line, a)
            self._rulerPts = dict()
            if not event.modifiers() & Qt.AltModifier:
                self._findIntersections()
            self.parent().update()
        else:
            super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event):
        if event.button() == Qt.LeftButton:
            # double click calls release twice
            if self._rulerObject is None:
                return
            line = self._rulerObject[0]
            if not line.length():
                # delete no-op ruler
                self.toolDisabled()
        else:
            super().mouseReleaseEvent(event)

    # custom painting

    def paint(self, painter, index):
        widget = self.parent()
        if index != widget.activeIndex():
            return
        scale = widget.inverseScale()
        # metrics
        if self._rulerObject is not None:
            line, a = self._rulerObject
            origin = line.p1()
            cursor = line.p2()
            size = 8 * scale
            halfSize = 4 * scale
            color = QColor(255, 85, 127, 170)

            # line
            painter.save()
            painter.setPen(color)
            drawing.drawLine(
                painter, origin.x(), origin.y(), cursor.x(), cursor.y(), scale
            )
            # ellipses
            ellipses = [(origin.x(), origin.y()), (cursor.x(), cursor.y())]
            path = QPainterPath()
            path.setFillRule(Qt.WindingFill)
            for x, y in itertools.chain(self._rulerPts.values(), ellipses):
                x -= halfSize
                y -= halfSize
                path.addEllipse(x, y, size, size)
            painter.fillPath(path, color)
            painter.restore()
            # text
            line = QLineF(line)
            xAlign = yAlign = "center"
            ellipses.pop(0)
            rp = self._rulerPts
            # XXX: sort shouldn't be performed in paintEvent
            for pt in itertools.chain((rp[k] for k in sorted(rp)), ellipses):
                p = QPointF(*pt)
                line.setP2(p)
                if line.length():
                    d = str(round(line.length(), 1))
                    pos = (line.p1() + line.p2()) / 2
                    drawing.drawTextAtPoint(
                        painter, d, pos.x(), pos.y(), scale, xAlign, yAlign
                    )
                line.setP1(p)
            xAlign, yAlign = "left", "top"
            dx = cursor.x() - origin.x()
            px = size
            if dx < 0:
                xAlign = "right"
                px = -px
            drawing.drawTextAtPoint(
                painter, a, cursor.x() + px, cursor.y() + size, scale, xAlign, yAlign
            )



"""
The *glyphViewFactory* submodule
-----------------------------

The *glyphViewFactory* submodule, as the name suggests, provides suitable
representations for the rendering of a Glyph_’s elements (points, Bézier
handles, components etc.).

.. _Glyph: http://ts-defcon.readthedocs.org/en/ufo3/objects/glyph.html
"""

import math

from fontTools.pens.basePen import BasePen
from fontTools.pens.qtPen import QtPen
from fontTools.pens.transformPen import TransformPen
from fontTools.ufoLib.pointPen import AbstractPointPen
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QPixmap
from PyQt5.QtWidgets import QGraphicsColorizeEffect

from defconQt.tools.drawing import applyEffectToPixmap, colorToQColor

# -------------
# no components
# -------------


def NoComponentsQPainterPathFactory(glyph):
    pen = NoComponentsQtPen(glyph.layer)
    glyph.draw(pen)
    pen.path.setFillRule(Qt.WindingFill)
    return pen.path


class NoComponentsQtPen(QtPen):
    def addComponent(self, glyphName, transformation):
        pass


# ---------------
# only components
# ---------------


def OnlyComponentsQPainterPathFactory(glyph):
    pen = OnlyComponentsQtPen(glyph.layer)
    glyph.draw(pen)
    pen.path.setFillRule(Qt.WindingFill)
    return pen.path


class OnlyComponentsQtPen(BasePen):
    def __init__(self, glyphSet):
        super().__init__(glyphSet)
        self.pen = QtPen(glyphSet)
        self.path = self.pen.path

    def _moveTo(self, p):
        pass

    def _lineTo(self, p):
        pass

    def _curveToOne(self, p1, p2, p3):
        pass

    def addComponent(self, glyphName, transformation):
        try:
            glyph = self.glyphSet[glyphName]
        except KeyError:
            return
        else:
            tPen = TransformPen(self.pen, transformation)
            glyph.draw(tPen)


# ----------
# point data
# ----------


def OutlineInformationFactory(glyph):
    pen = OutlineInformationPen()
    glyph.drawPoints(pen)
    return pen.getData()


class OutlineInformationPen(AbstractPointPen):
    def __init__(self):
        self._rawPointData = []
        self._rawComponentData = []
        self._bezierHandleData = []

    def getData(self):
        data = dict(
            onCurvePoints=[],
            offCurvePoints=[],
            bezierHandles=[],
            anchors=[],
            components=self._rawComponentData,
        )
        for contour in self._rawPointData:
            # anchor
            # TODO: UFO3 doesn't do this?
            if len(contour) == 1 and contour[0]["name"] is not None:
                anchor = contour[0]
                data["anchors"].append(anchor)
            # points
            else:
                for pointIndex, point in enumerate(contour):
                    if point["segmentType"] is None:
                        data["offCurvePoints"].append(point)
                        # look for handles
                        back = contour[pointIndex - 1]
                        forward = contour[(pointIndex + 1) % len(contour)]
                        if back["segmentType"] is not None:
                            p1 = back["point"]
                            p2 = point["point"]
                            if p1 != p2:
                                data["bezierHandles"].append(p1 + p2)
                            # only allow two handles a point for qcurve
                            if forward["segmentType"] != "qcurve":
                                continue
                        if forward["segmentType"] is not None:
                            p1 = forward["point"]
                            p2 = point["point"]
                            if p1 != p2:
                                data["bezierHandles"].append(p1 + p2)
                    else:
                        # catch first point
                        if not pointIndex or point["smooth"]:
                            nextOn = None
                            for nextPoint in (
                                contour[pointIndex:] + contour[:pointIndex]
                            ):
                                if nextPoint["point"] == point["point"]:
                                    continue
                                nextOn = nextPoint
                                break
                            angle = None
                            if nextOn:
                                x1, y1 = point["point"]
                                x2, y2 = nextOn["point"]
                                xDiff = x2 - x1
                                yDiff = y2 - y1
                                angle = math.atan2(yDiff, xDiff)
                            # store
                            if not pointIndex:
                                point["startPointAngle"] = angle
                            if point["smooth"]:
                                # no point storing None angle here
                                if angle is not None:
                                    angle -= 0.5 * math.pi
                                    point["smoothAngle"] = angle
                        data["onCurvePoints"].append(point)
        return data

    def beginPath(self):
        self._rawPointData.append([])

    def endPath(self):
        pass

    def addPoint(self, pt, segmentType=None, smooth=False, name=None, **kwargs):
        d = dict(point=pt, segmentType=segmentType, smooth=smooth, name=name, **kwargs)
        self._rawPointData[-1].append(d)

    def addComponent(self, baseGlyphName, transformation):
        self._rawComponentData.append((baseGlyphName, transformation))


# -----
# image
# -----


def QPixmapFactory(image):
    font = image.font
    if font is None:
        return None
    _ = image.layer
    images = font.images
    if image.fileName not in images:
        return None
    imageColor = image.color
    data = images[image.fileName]
    pixmap = QPixmap()
    pixmap.loadFromData(data)
    if imageColor is not None:
        colorEffect = QGraphicsColorizeEffect()
        colorEffect.setColor(colorToQColor(imageColor))
        colorEffect.setStrength(0.8)
        return applyEffectToPixmap(pixmap, colorEffect)
    return pixmap



from PyQt5.QtCore import QSize, Qt
from PyQt5.QtGui import QPainter
from PyQt5.QtWidgets import QAbstractButton


class PathButton(QAbstractButton):
    def __init__(self, parent=None):
        super().__init__(parent)
        # TODO: make it TabFocus + make sure there's a corresponding visual cue
        self.setFocusPolicy(Qt.NoFocus)

        self._drawingCommands = []
        self._isDownColor = None
        self._isFlipped = False
        self._size = QSize(28, 28)

    def drawingCommands(self):
        return self._drawingCommands

    def setDrawingCommands(self, commands):
        self._drawingCommands = commands
        self.update()

    def isFlipped(self):
        return self._isFlipped

    def setIsFlipped(self, value):
        self._isFlipped = value

    def isDownColor(self):
        return self._isDownColor

    def setIsDownColor(self, color):
        self._isDownColor = color
        self.update()

    def size(self):
        return self._size

    def setSize(self, size):
        self._size = size
        self.adjustSize()

    # ----------
    # Qt methods
    # ----------

    def paintEvent(self, event):
        if self._drawingCommands is None:
            return
        painter = QPainter(self)
        rect = event.rect()
        target = self.rect().size()
        size = self._drawingCommands[0]
        if not size.isNull() and (
            size.width() < target.width() or size.height() < target.height()
        ):
            dx = dy = 0
            if size.width() < target.width():
                dx = int(0.5 * (target.width() - size.width()))
            if size.height() < target.height():
                dy = int(0.5 * (target.height() - size.height()))
            painter.translate(dx, dy)
        isDown = self.isDown() and self._isDownColor is not None
        if isDown and self._isDownColor.isValid():
            painter.fillRect(rect, self._isDownColor)
            isDown = False
        if self._isFlipped:
            painter.translate(0, target.height())
            painter.scale(1, -1)
        for path, cmd, color in self._drawingCommands[1:]:
            if cmd == "f":
                if isDown:
                    color = color.darker(120)
                painter.save()
                painter.setRenderHint(QPainter.Antialiasing)
                painter.fillPath(path, color)
                painter.restore()
            else:
                if isDown:
                    color = color.darker(150)
                painter.save()
                if cmd[-1] == "a":
                    painter.setRenderHint(QPainter.Antialiasing)
                pen = painter.pen()
                pen.setColor(color)
                pen.setWidth(int(cmd[0]))
                painter.setPen(pen)
                painter.drawPath(path)
                painter.restore()

    def minimumSizeHint(self):
        return self.sizeHint()

    def sizeHint(self):
        return self._size



from PyQt5.QtCore import QSize, pyqtSignal
from PyQt5.QtGui import QColor, QKeySequence, QPainter
from PyQt5.QtWidgets import QSizePolicy, QVBoxLayout, QWidget

from trufont.controls.pathButton import PathButton


class ToolBar(QWidget):
    """
    TODO: allow all orientations
    """

    currentToolChanged = pyqtSignal(object)

    def __init__(self, parent=None):
        super().__init__(parent)
        self._color = QColor(90, 90, 90)
        self._selectedColor = QColor(20, 146, 230)
        self._currentTool = 0
        self._tools = []

        self.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Expanding)
        layout = QVBoxLayout(self)
        layout.setContentsMargins(8, 12, 8, 12)
        layout.setSpacing(12)

    def _buttonClicked(self):
        layout = self.layout()
        sender = self.sender()
        toolIndex = 0
        for i in range(layout.count() - 1):
            btn = layout.itemAt(i).widget()
            if btn == sender:
                color = self._selectedColor
                toolIndex = i
            else:
                color = self._color
            cmds = btn.drawingCommands()
            cmds[1][2] = color
            btn.setDrawingCommands(cmds)
        self._currentTool = toolIndex
        self.currentToolChanged.emit(self._tools[toolIndex])

    def _updateColors(self):
        layout = self.layout()
        for i in range(layout.count() - 1):
            btn = layout.itemAt(i).widget()
            if self._currentTool == i:
                color = self._selectedColor
            else:
                color = self._color
            cmds = btn.drawingCommands()
            cmds[1][2] = color
            btn.setDrawingCommands(cmds)

    def _updateTools(self):
        layout = self.layout()
        for i in reversed(range(layout.count())):
            layout.takeAt(i).widget().close()
        for index, tool in enumerate(self._tools):
            btn = PathButton(self)
            if index == self._currentTool:
                color = self._selectedColor
            else:
                color = self._color
            btn.setDrawingCommands([QSize(28, 28), [tool.icon, "f", color]])
            btn.setIsDownColor(QColor())
            btn.setIsFlipped(True)
            text = tool.name
            if tool.shortcut is not None:
                text = f"{text} ({tool.shortcut})"
                btn.setShortcut(QKeySequence(tool.shortcut))
            btn.setToolTip(text)
            btn.clicked.connect(self._buttonClicked)
            layout.addWidget(btn)
        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding)
        layout.addWidget(spacer)

    def color(self):
        return self._color

    def setColor(self, color):
        self._color = color
        self._updateColors()

    def selectedColor(self):
        return self._color

    def setSelectedColor(self, color):
        self._selectedColor = color
        self._updateColors()

    def tools(self):
        return self._tools

    def setTools(self, tools):
        self._tools = list(tools)
        self._updateTools()

    def addTool(self, tool):
        self._tools.append(tool)
        self._updateTools()

    def removeTool(self, tool):
        self._tools.remove(tool)
        self._updateTools()

    def currentTool(self):
        return self._tools[self._currentTool]

    def setCurrentTool(self, tool):
        self._currentTool = self._tools.index(tool)
        self._updateColors()

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.fillRect(event.rect(), QColor(240, 240, 240))



from PyQt5.QtCore import QObject

from defconQt.controls.glyphContextView import GlyphRecord


def _reverseEnumerate(seq):
    n = len(seq)
    for obj in reversed(seq):
        n -= 1
        yield n, obj


# TODO: support unencoded glyphs input
# TODO: support RTL (caret)
# TODO: support selection


class LayoutManager(QObject):
    """
    parent should be a GlyphContextView.
    """

    def __init__(self, font, parent):
        super().__init__(parent)
        self._font = font
        # this stays None except when we want to commit to widget
        self._activeIndex = None
        self._caretIndex = 0
        self._glyphList = []
        self._needsCaretPostFix = False
        self._needsLayout = False
        self.updateView()

    def glyphList(self):
        return self._glyphList

    def setGlyphList(self, glyphNames):
        self._glyphList = glyphNames
        self._caretIndex = len(self._glyphList)
        self._needsLayout = True
        self.updateView()

    #

    def initCaret(self):
        """
        aka. setCaretFromActiveIndex
        """
        widget = self.parent()
        index = widget.activeIndex()
        glyphRecords = widget.glyphRecords()
        try:
            self._caretIndex = glyphRecords[index + 1].cluster
        except IndexError:
            self._caretIndex = len(self._glyphList)

    def caretNext(self):
        widget = self.parent()
        # TODO: this could be a binary search
        for i, rec in enumerate(widget.glyphRecords()):
            if rec.cluster > self._caretIndex:
                self._activeIndex = i - 1
                self._caretIndex = rec.cluster
                break
        # TODO: trim no-op updates?
        if self._activeIndex is None:
            self._activeIndex = i
            self._caretIndex = len(self._glyphList)
        self.updateView()

    def caretPrevious(self):
        widget = self.parent()
        # TODO: this could be a binary search
        for i, rec in _reverseEnumerate(widget.glyphRecords()):
            if rec.cluster < self._caretIndex:
                self._activeIndex = max(i - 1, 0)
                self._caretIndex = rec.cluster
                break
        if self._activeIndex is None:
            # we're at left boundary
            return
        self.updateView()

    def setCaretFromPos(self, pos):
        """
        *pos* is provided in parent coordinates.
        """
        widget = self.parent()
        index = widget.indexForPoint(pos)
        if index is not None:
            glyphRecords = widget.glyphRecords()
            record = glyphRecords[index]
            pos.setX(pos.x() + 0.5 * (record.xAdvance * widget.scale()))
            halfIndex = widget.indexForPoint(pos)
            if halfIndex == index:
                self._activeIndex = max(index - 1, 0)
                self._caretIndex = record.cluster
            else:
                self._activeIndex = index
                try:
                    self._caretIndex = glyphRecords[index + 1].cluster
                except IndexError:
                    self._caretIndex = len(self._glyphList)
            self.updateView()

    def insert(self, content):
        if isinstance(content, list):
            self._glyphList[self._caretIndex : self._caretIndex] = content
            self._caretIndex += len(content)
        else:
            self._glyphList.insert(self._caretIndex, content)
            self._caretIndex += 1
        # clamp caretIndex after shaping and set activeIndex
        # typ. when we input text that serves as prefix to a ligature
        self._needsCaretPostFix = True
        self._needsLayout = True
        self.updateView()

    def delete(self, forward=False):
        index = self._caretIndex - (not forward)
        if index < 0 or index >= len(self._glyphList):
            return
        self._glyphList.pop(index)
        self._caretIndex = index
        self._needsCaretPostFix = True
        self._needsLayout = True
        self.updateView()

    #

    def drawingOffset(self, index):
        """
        Returns *(dx, dy)* offset if caret is to be drawn, or None otherwise.
        """
        glyphRecords = self.parent().glyphRecords()
        glyphRecord = glyphRecords[index]
        atRightBoundary = self._caretIndex == len(self._glyphList)
        if glyphRecord.cluster != self._caretIndex and not (
            atRightBoundary and index == len(glyphRecords) - 1
        ):
            return None
        dx = -glyphRecord.xOffset
        dy = -glyphRecord.yOffset
        if atRightBoundary:
            dx += glyphRecord.xAdvance
            dy += glyphRecord.yAdvance
        return dx, dy

    #

    @property
    def _shaper(self):
        engine = self._font.engine
        if engine is None:
            return None
        if hasattr(engine, "_layoutEngine"):
            return "compositor"
        return "harfbuzz"

    def _shapeAndSetText(self):
        font = self._font
        records = font.engine.process(self._glyphList)
        if self._shaper == "compositor":
            records_ = []
            index = 0
            for glyphRecord in records:
                record_ = GlyphRecord()
                record_.glyph = glyph = font[glyphRecord.glyphName]
                record_.cluster = index
                record_.xOffset = glyphRecord.xPlacement
                record_.yOffset = glyphRecord.yPlacement
                record_.xAdvance = glyph.width + glyphRecord.xAdvance
                record_.yAdvance = glyph.height + glyphRecord.yAdvance
                records_.append(record_)
                index += len(glyphRecord.ligatureComponents) or 1
            records = records_
        self.parent().setGlyphRecords(records)

    def updateView(self):
        widget = self.parent()
        if self._needsLayout:
            font = self._font
            if font.engine is not None:
                self._shapeAndSetText()
            else:
                glyphs = []
                for glyphName in self._glyphList:
                    if glyphName in font:
                        glyphs.append(font[glyphName])
                widget.setGlyphs(glyphs)
            self._needsLayout = False
        if self._needsCaretPostFix:
            glyphRecords = widget.glyphRecords()
            for i, rec in enumerate(glyphRecords):
                if rec.cluster >= self._caretIndex:
                    self._activeIndex = max(i - 1, 0)
                    self._caretIndex = rec.cluster
                    break
            if self._activeIndex is None:
                self._activeIndex = max(len(glyphRecords) - 1, 0)
            self._needsCaretPostFix = False
        if self._activeIndex is not None:
            widget.setActiveIndex(self._activeIndex)
            self._activeIndex = None
        widget.update()



from PyQt5.QtCore import Qt
from PyQt5.QtGui import QColor, QFontMetrics, QPainter, QPainterPath, QPixmap

from defconQt.tools import platformSpecific
from defconQt.tools.drawing import colorToQColor

GlyphCellHeaderHeight = 13
GlyphCellMinHeightForHeader = 40
GlyphCellMinHeightForMetrics = 100

cellMetricsFillColor = cellMetricsLineColor = QColor(224, 226, 220)
cellMetricsTextColor = QColor(72, 72, 72)
cellDirtyColor = QColor(240, 240, 240, 170)

headerFont = platformSpecific.otherUIFont()

# TODO: fine-tune dirty appearance


def GlyphCellFactory(
    glyph,
    width,
    height,
    drawLayers=False,
    drawMarkColor=True,
    drawHeader=None,
    drawMetrics=None,
    pixelRatio=1.0,
):
    if drawHeader is None:
        drawHeader = height >= GlyphCellMinHeightForHeader
    if drawMetrics is None:
        drawMetrics = height >= GlyphCellMinHeightForMetrics
    obj = GlyphCellFactoryDrawingController(
        glyph=glyph,
        font=glyph.font,
        width=width,
        height=height,
        drawLayers=False,
        drawMarkColor=drawMarkColor,
        drawHeader=drawHeader,
        drawMetrics=drawMetrics,
        pixelRatio=pixelRatio,
    )
    return obj.getPixmap()


class GlyphCellFactoryDrawingController:
    """
    This draws the cell with the layers stacked in this order:

    ------------------
    header text
    ------------------
    header background
    ------------------
    foreground
    ------------------
    glyph
    ------------------
    vertical metrics
    ------------------
    horizontal metrics
    ------------------
    background
    ------------------

    Subclasses may override the layer drawing methods to customize
    the appearance of cells.
    """

    def __init__(
        self,
        glyph,
        font,
        width,
        height,
        pixelRatio=1.0,
        drawLayers=False,
        drawMarkColor=True,
        drawHeader=True,
        drawMetrics=True,
    ):
        self.glyph = glyph
        self.font = font
        self.pixelRatio = pixelRatio
        self.width = width
        self.height = height
        self.bufferPercent = 0.10
        self.shouldDrawHeader = drawHeader
        self.shouldDrawLayers = drawLayers
        self.shouldDrawMarkColor = drawMarkColor
        self.shouldDrawMetrics = drawMetrics

        self.headerAtBottom = True
        self.headerHeight = 0
        if drawHeader:
            self.headerHeight = GlyphCellHeaderHeight
        availableHeight = (height - self.headerHeight) * (
            1.0 - (self.bufferPercent * 2)
        )
        descender = font.info.descender or -250
        # some fonts overflow their upm, try to infer
        if font.info.descender and font.info.ascender:
            unitsPerEm = font.info.ascender - descender
        else:
            unitsPerEm = font.info.unitsPerEm or 1000
        self.buffer = height * self.bufferPercent
        self.scale = availableHeight / unitsPerEm
        self.xOffset = (width - (glyph.width * self.scale)) / 2
        self.yOffset = abs(descender * self.scale) + 0.4 * self.buffer

    def getPixmap(self):
        pixmap = QPixmap(self.width * self.pixelRatio, self.height * self.pixelRatio)
        pixmap.setDevicePixelRatio(self.pixelRatio)
        pixmap.fill(Qt.transparent)
        painter = QPainter(pixmap)
        painter.setRenderHint(QPainter.Antialiasing)
        painter.translate(0, self.height)
        painter.scale(1, -1)
        if self.headerAtBottom:
            bodyRect = (0, 0, self.width, self.height - self.headerHeight)
            headerRect = (0, 0, self.width, self.headerHeight)
        else:
            bodyRect = (0, 0, self.width, self.height - self.headerHeight)
            headerRect = (0, 0, self.width, self.headerHeight)
        # background
        painter.save()
        if self.headerAtBottom:
            h = self.height
        else:
            h = self.height - self.headerHeight
        painter.translate(0, h)
        painter.scale(1, -1)
        self.drawCellBackground(painter, bodyRect)
        painter.restore()
        # glyph
        if self.headerAtBottom:
            painter.translate(0, self.headerHeight)
        if self.shouldDrawMetrics:
            self.drawCellHorizontalMetrics(painter, bodyRect)
            self.drawCellVerticalMetrics(painter, bodyRect)
        painter.save()
        painter.setClipRect(0, 0, self.width, self.height - self.headerHeight)
        painter.translate(self.xOffset, self.yOffset)
        painter.scale(self.scale, self.scale)
        self.drawCellGlyph(painter)
        painter.restore()
        # foreground
        painter.save()
        painter.translate(0, self.height - self.headerHeight)
        painter.scale(1, -1)
        self.drawCellForeground(painter, bodyRect)
        painter.restore()
        # header
        if self.shouldDrawHeader:
            painter.save()
            if self.headerAtBottom:
                h = 0
            else:
                h = self.height
            painter.translate(0, h)
            painter.scale(1, -1)
            self.drawCellHeaderBackground(painter, headerRect)
            self.drawCellHeaderText(painter, headerRect)
            painter.restore()
        return pixmap

    def drawCellBackground(self, painter, rect):
        if self.shouldDrawMarkColor:
            markColor = self.glyph.markColor
            if markColor is not None:
                color = colorToQColor(markColor)
                color.setAlphaF(0.7 * color.alphaF())
                painter.fillRect(*(rect + (color,)))
        if self.shouldDrawHeader:
            if self.glyph.dirty:
                x, _, w, _ = rect
                painter.fillRect(*(rect + (cellDirtyColor,)))
                path = QPainterPath()
                path.moveTo(x + w - 12, 0)
                path.lineTo(x + w, 0)
                path.lineTo(x + w, 12)
                path.closeSubpath()
                painter.fillPath(path, QColor(255, 0, 0, 170))

    def drawCellHorizontalMetrics(self, painter, rect):
        xMin, yMin, _, _ = rect
        glyph = self.glyph
        scale = self.scale
        xOffset = self.xOffset
        left = round((0 * scale) + xMin + xOffset)
        right = round((glyph.width * scale) + xMin + xOffset)

        hi = 750
        lo = -250
        font = self.font
        yOffset = self.yOffset
        if font is not None:
            ascender = font.info.ascender or 750
            capHeight = font.info.capHeight or 750
            hi = max(ascender, capHeight)
            lo = font.info.descender or -250
        hi = round((hi * scale) + yMin + yOffset)
        lo = round((lo * scale) + yMin + yOffset)

        painter.save()
        painter.setPen(cellMetricsFillColor)
        painter.setRenderHint(QPainter.Antialiasing, False)
        painter.drawLine(left, lo, left, hi)
        painter.drawLine(right, lo, right, hi)
        painter.restore()

    def drawCellVerticalMetrics(self, painter, rect):
        xMin, yMin, _, _ = rect
        font = self.font
        scale = self.scale
        xOffset, yOffset = self.xOffset, self.yOffset
        left = round((0 * scale) + xMin + xOffset)
        right = round((self.glyph.width * scale) + xMin + xOffset)
        lines = {
            0,
            font.info.descender,
            font.info.xHeight,
            font.info.capHeight,
            font.info.ascender,
        }
        painter.save()
        painter.setPen(cellMetricsLineColor)
        painter.setRenderHint(QPainter.Antialiasing, False)
        for y in lines:
            if y is None:
                continue
            y = round((y * scale) + yMin + yOffset)
            painter.drawLine(left, y, right, y)
        painter.restore()

    def drawCellGlyph(self, painter):
        if self.shouldDrawLayers:
            layers = self.font.layers
            for layerName in reversed(layers.layerOrder):
                layer = layers[layerName]
                if self.glyph.name not in layer:
                    continue
                layerColor = None
                if layer.color is not None:
                    layerColor = colorToQColor(layer.color)
                if layerColor is None:
                    layerColor = Qt.black
                glyph = layer[self.glyph.name]
                path = glyph.getRepresentation("defconQt.QPainterPath")
                painter.fillPath(path, layerColor)
        else:
            path = self.glyph.getRepresentation("defconQt.QPainterPath")
            painter.fillPath(path, Qt.black)

    def drawCellForeground(self, painter, rect):
        pass

    def drawCellHeaderBackground(self, painter, rect):
        xMin, yMin, width, height = rect
        # background
        if self.shouldDrawMarkColor and self.glyph.markColor is not None:
            color = colorToQColor(self.glyph.markColor)
        elif self.glyph.dirty:
            color = cellDirtyColor
        else:
            color = Qt.white
        painter.fillRect(xMin, yMin, width, height, color)

    def drawCellHeaderText(self, painter, rect):
        _, _, width, height = rect
        metrics = QFontMetrics(headerFont)
        minOffset = painter.pen().width()

        painter.setFont(headerFont)
        painter.setPen(cellMetricsTextColor)
        name = metrics.elidedText(self.glyph.name, Qt.ElideRight, width - 2)
        painter.drawText(
            1,
            0,
            width - 2,
            height - minOffset,
            Qt.TextSingleLine | Qt.AlignCenter | Qt.AlignBottom,
            name,
        )



import sys
import unittest

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QWidget

from trufont.controls.nameTabWidget import NameTabWidget
from trufont.objects.application import Application
from trufont.objects.defcon import TFont
from trufont.windows.fontWindow import FontWindow


class NameTabWidgetTest(unittest.TestCase):

    app = Application(sys.argv)

    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)

    def setUp(self):
        self.infoTab = NameTabWidget(None)

    def test_addNamedTab(self):
        tab1 = QWidget(None)
        tab1.name = "Test1"
        self.infoTab.addNamedTab(tab1)
        tab2 = self.infoTab.widget(0)
        self.assertEqual(tab1, tab2)
        self.assertEqual(tab1.name, tab2.name)

        tab2 = QWidget(None)
        tab2.name = "Test2"
        self.infoTab.addNamedTab(tab2)
        tab2 = self.infoTab.widget(0)
        self.assertEqual(tab1, tab2)
        self.assertEqual(tab1.name, tab2.name)


# TODO TabDialogTest

# TODO TabWidget


class TabTestCase(unittest.TestCase):

    app = Application(sys.argv)

    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)

    def setUp(self):
        self.font = TFont.new()

        self.mainWindow = FontWindow(self.font)
        self.mainWindow.fontInfo()
        self.fontInfo = self.mainWindow._infoWindow
        self.generalTab = self.fontInfo.tabWidget.widget(0)
        self.legalTab = self.fontInfo.tabWidget.widget(1)
        self.openTypeTab = self.fontInfo.tabWidget.widget(2)
        self.postScriptTab = self.fontInfo.tabWidget.widget(3)
        self.fontInfo.show()

    def tearDown(self):
        self.fontInfo.close()

    def checkString(self, attrName):
        attrEdit = getattr(self.tab, attrName + "Edit")
        attrEdit.setEnabled(True)

        self.assertIsNone(getattr(self.font.info, attrName))
        self.assertEqual(attrEdit.text(), "")

        attrEdit.setText("Typeface " + attrName)
        self.fontInfo.accept()
        self.assertEqual(getattr(self.font.info, attrName), "Typeface " + attrName)

        attrEdit.setEnabled(False)
        self.fontInfo.accept()
        self.assertIsNone(getattr(self.font.info, attrName))

    def checkMultilineString(self, attrName):
        attrEdit = getattr(self.tab, attrName + "Edit")
        attrEdit.setEnabled(True)

        self.assertIsNone(getattr(self.font.info, attrName))
        self.assertEqual(attrEdit.toPlainText(), "")

        attrEdit.setPlainText("Typeface \n" + attrName)
        self.fontInfo.accept()
        self.assertEqual(getattr(self.font.info, attrName), "Typeface \n" + attrName)

        attrEdit.setEnabled(False)
        self.fontInfo.accept()
        self.assertIsNone(getattr(self.font.info, attrName))

    def checkInteger(self, attrName, value=0):
        attrEdit = getattr(self.tab, attrName + "Edit")
        attrEdit.setEnabled(True)

        if value == 0:
            self.assertIsNone(getattr(self.font.info, attrName))
            self.assertEqual(attrEdit.value(), 0)
        else:
            self.assertIsNotNone(getattr(self.font.info, attrName))
            self.assertEqual(attrEdit.value(), value)

        attrEdit.setValue(123)
        self.fontInfo.accept()
        self.assertEqual(getattr(self.font.info, attrName), 123)

        attrEdit.setValue(0)
        self.fontInfo.accept()
        self.assertEqual(getattr(self.font.info, attrName), 0)

        attrEdit.setValue(-123)
        self.fontInfo.accept()
        self.assertEqual(getattr(self.font.info, attrName), -123)

        attrEdit.setEnabled(False)
        self.fontInfo.accept()
        self.assertIsNone(getattr(self.font.info, attrName))

        self.assertEqual(attrEdit.minimum(), -2147483648)
        self.assertEqual(attrEdit.maximum(), 2147483647)

    def checkPositiveInteger(self, attrName, value=0):
        attrEdit = getattr(self.tab, attrName + "Edit")
        attrEdit.setEnabled(True)

        if value == 0:
            self.assertIsNone(getattr(self.font.info, attrName))
            self.assertEqual(attrEdit.value(), 0)
        else:
            self.assertIsNotNone(getattr(self.font.info, attrName))
            self.assertEqual(attrEdit.value(), value)

        attrEdit.setValue(123)
        self.fontInfo.accept()
        self.assertEqual(getattr(self.font.info, attrName), 123)

        attrEdit.setValue(0)
        self.fontInfo.accept()
        self.assertEqual(getattr(self.font.info, attrName), 0)

        attrEdit.setEnabled(False)
        self.fontInfo.accept()
        self.assertIsNone(getattr(self.font.info, attrName))

        self.assertEqual(attrEdit.minimum(), 0)
        self.assertEqual(attrEdit.maximum(), 2147483647)

    def checkIntegerFloat(self, attrName, value=0):
        attrEdit = getattr(self.tab, attrName + "Edit")
        attrEdit.setEnabled(True)

        if value == 0:
            self.assertIsNone(getattr(self.font.info, attrName))
            self.assertEqual(attrEdit.value(), 0)
        else:
            self.assertIsNotNone(getattr(self.font.info, attrName))
            self.assertEqual(attrEdit.value(), value)

        attrEdit.setValue(123)
        self.fontInfo.accept()
        attr = getattr(self.font.info, attrName)
        self.assertTrue(isinstance(attr, int))
        self.assertEqual(attr, 123)

        attrEdit.setValue(123.000)
        self.fontInfo.accept()
        attr = getattr(self.font.info, attrName)
        self.assertTrue(isinstance(attr, int))
        self.assertEqual(attr, 123)

        attrEdit.setValue(123.456)
        self.fontInfo.accept()
        attr = getattr(self.font.info, attrName)
        self.assertTrue(isinstance(attr, float))
        self.assertEqual(attr, 123.456)

        attrEdit.setValue(0)
        self.fontInfo.accept()
        attr = getattr(self.font.info, attrName)
        self.assertTrue(isinstance(attr, int))
        self.assertEqual(attr, 0)

        attrEdit.setValue(0.0)
        self.fontInfo.accept()
        attr = getattr(self.font.info, attrName)
        self.assertTrue(isinstance(attr, int))
        self.assertEqual(attr, 0)

        attrEdit.setValue(-123)
        self.fontInfo.accept()
        attr = getattr(self.font.info, attrName)
        self.assertTrue(isinstance(attr, int))
        self.assertEqual(attr, -123)

        attrEdit.setValue(-123.000)
        self.fontInfo.accept()
        attr = getattr(self.font.info, attrName)
        self.assertTrue(isinstance(attr, int))
        self.assertEqual(attr, -123)

        attrEdit.setValue(-123.456)
        self.fontInfo.accept()
        attr = getattr(self.font.info, attrName)
        self.assertTrue(isinstance(attr, float))
        self.assertEqual(attr, -123.456)

        attrEdit.setEnabled(False)
        self.fontInfo.accept()
        self.assertIsNone(getattr(self.font.info, attrName))

        self.assertEqual(attrEdit.minimum(), -2147483648)
        self.assertEqual(attrEdit.maximum(), 2147483647)

    def checkPositiveIntegerFloat(self, attrName, value=0):
        attrEdit = getattr(self.tab, attrName + "Edit")
        attrEdit.setEnabled(True)

        if value == 0:
            self.assertIsNone(getattr(self.font.info, attrName))
            self.assertEqual(attrEdit.value(), 0)
        else:
            self.assertIsNotNone(getattr(self.font.info, attrName))
            self.assertEqual(attrEdit.value(), value)

        attrEdit.setValue(123)
        self.fontInfo.accept()
        attr = getattr(self.font.info, attrName)
        self.assertTrue(isinstance(attr, int))
        self.assertEqual(attr, 123)

        attrEdit.setValue(123.000)
        self.fontInfo.accept()
        attr = getattr(self.font.info, attrName)
        self.assertTrue(isinstance(attr, int))
        self.assertEqual(attr, 123)

        attrEdit.setValue(123.456)
        self.fontInfo.accept()
        attr = getattr(self.font.info, attrName)
        self.assertTrue(isinstance(attr, float))
        self.assertEqual(attr, 123.456)

        attrEdit.setValue(0)
        self.fontInfo.accept()
        attr = getattr(self.font.info, attrName)
        self.assertTrue(isinstance(attr, int))
        self.assertEqual(attr, 0)

        attrEdit.setValue(0.0)
        self.fontInfo.accept()
        attr = getattr(self.font.info, attrName)
        self.assertTrue(isinstance(attr, int))
        self.assertEqual(attr, 0)

        attrEdit.setEnabled(False)
        self.fontInfo.accept()
        self.assertIsNone(getattr(self.font.info, attrName))

        self.assertEqual(attrEdit.minimum(), 0)
        self.assertEqual(attrEdit.maximum(), 2147483647)

    # XXX: Implement maxLen, evenLen?
    def checkIntegerFloatList(self, attrName, maxLen=None, evenLen=None):
        attrEdit = getattr(self.tab, attrName + "Edit")
        attrEdit.setEnabled(True)

        attrEdit.setText("")
        self.fontInfo.accept()
        self.assertEqual(getattr(self.font.info, attrName), [])

        attrEdit.setText("123 456 789 0")
        self.fontInfo.accept()
        attr = getattr(self.font.info, attrName)
        self.assertEqual(attr, [123, 456, 789, 0])

        attrEdit.setText("123,0 456,1 789.11111112 789,11111113 790 791")
        self.fontInfo.accept()
        attr = getattr(self.font.info, attrName)
        self.assertEqual(attr, [123, 456.1, 789.11111112, 789.11111113, 790, 791])

        attrEdit.setEnabled(False)
        self.fontInfo.accept()
        attr = getattr(self.font.info, attrName)
        # These are apparently always [], never None.
        if attrName in [
            "postscriptBlueValues",
            "postscriptOtherBlues",
            "postscriptFamilyBlues",
            "postscriptFamilyOtherBlues",
            "postscriptStemSnapH",
            "postscriptStemSnapV",
        ]:
            self.assertEqual(attr, [])
        else:
            self.assertIsNone(attr)

    def checkBoolean(self, attrName):
        attrEdit = getattr(self.tab, attrName + "Edit")
        attrEdit.setEnabled(True)

        # Most checkboxes are governed by a parent checkbox for the field (see
        # Postscript tab for examples), meaning that if the field is unchecked,
        # it is assumed that the user doesn't want to set an explicit value. If
        # the field is checked, the actual checkbox can be either checked or
        # not.
        self.assertFalse(attrEdit.isTristate())

        attrEdit.setCheckState(Qt.Checked)
        self.fontInfo.accept()
        value = attrEdit.checkState()
        self.assertEqual(value, Qt.Checked)
        self.assertEqual(getattr(self.font.info, attrName), True)

        attrEdit.setCheckState(Qt.Unchecked)
        self.fontInfo.accept()
        value = attrEdit.checkState()
        self.assertEqual(value, Qt.Unchecked)
        self.assertEqual(getattr(self.font.info, attrName), False)

        attrEdit.setEnabled(False)
        self.fontInfo.accept()
        self.assertIsNone(getattr(self.font.info, attrName))


class GeneralTabTest(TabTestCase):
    def __init__(self, methodName):
        super().__init__(methodName)

    def setUp(self):
        super().setUp()
        self.tab = self.generalTab

    def test_GeneralTab_name(self):
        self.assertEqual(self.tab.name, "General")

    def test_familyName(self):
        self.checkString("familyName")

    def test_styleName(self):
        self.checkString("styleName")

    # TODO styleMapFamilyName, ...DateCreated

    def test_versionMajor(self):
        self.checkInteger("versionMajor")

    def test_versionMinor(self):
        self.checkPositiveInteger("versionMinor")

    def test_unitsPerEm(self):
        self.checkPositiveIntegerFloat("unitsPerEm", value=1000)

    def test_italicAngle(self):
        self.checkIntegerFloat("italicAngle")

    def test_ascender(self):
        self.checkIntegerFloat("ascender", 750)

    def test_descender(self):
        self.checkIntegerFloat("descender", -250)

    def test_capHeight(self):
        self.checkIntegerFloat("capHeight", 700)

    def test_xHeight(self):
        self.checkIntegerFloat("xHeight", 500)

    def test_noteEdit(self):
        self.checkMultilineString("note")


class LegalTabTest(TabTestCase):
    def __init__(self, methodName):
        super().__init__(methodName)

    def setUp(self):
        super().setUp()
        self.tab = self.legalTab

    def test_MetricsTab_name(self):
        self.assertEqual(self.tab.name, "Legal")

    def test_designer(self):
        self.checkString("openTypeNameDesigner")

    def test_designerURL(self):
        self.checkString("openTypeNameDesignerURL")

    def test_manufacturer(self):
        self.checkString("openTypeNameManufacturer")

    def test_manufacturerURL(self):
        self.checkString("openTypeNameManufacturerURL")

    def test_copyright(self):
        self.checkMultilineString("copyright")

    def test_license(self):
        self.checkMultilineString("openTypeNameLicense")

    def test_licenseURL(self):
        self.checkString("openTypeNameLicenseURL")

    def test_trademark(self):
        self.checkString("trademark")


class OpenTypeTabTest(TabTestCase):
    def __init__(self, methodName):
        super().__init__(methodName)

    def setUp(self):
        super().setUp()
        self.tab = self.openTypeTab

    def test_OpenTypeTab_name(self):
        self.assertEqual(self.tab.name, "OpenType")

    def test_lowestRecPPEM(self):
        self.checkPositiveInteger("openTypeHeadLowestRecPPEM")

    # TODO: head flags

    def test_preferredFamilyName(self):
        self.checkString("openTypeNamePreferredFamilyName")

    def test_preferredSubfamilyName(self):
        self.checkString("openTypeNamePreferredSubfamilyName")

    def test_compatibleFullName(self):
        self.checkString("openTypeNameCompatibleFullName")

    def test_WWSFamilyName(self):
        self.checkString("openTypeNameWWSFamilyName")

    def test_WWSSubfamilyName(self):
        self.checkString("openTypeNameWWSSubfamilyName")

    def test_version(self):
        self.checkString("openTypeNameVersion")

    def test_uniqueID(self):
        self.checkString("openTypeNameUniqueID")

    def test_description(self):
        self.checkString("openTypeNameDescription")

    def test_sampleText(self):
        self.checkString("openTypeNameSampleText")

    # TODO: name records table

    def test_ascender(self):
        self.checkInteger("openTypeHheaAscender")

    def test_descender(self):
        self.checkInteger("openTypeHheaDescender")

    def test_lineGap(self):
        self.checkInteger("openTypeHheaLineGap")

    def test_caretSlopeRise(self):
        self.checkInteger("openTypeHheaCaretSlopeRise")

    def test_caretSlopeRun(self):
        self.checkInteger("openTypeHheaCaretSlopeRun")

    def test_caretOffset(self):
        self.checkInteger("openTypeHheaCaretOffset")

    def test_vertTypoAscender(self):
        self.checkInteger("openTypeVheaVertTypoAscender")

    def test_vertTypoDescender(self):
        self.checkInteger("openTypeVheaVertTypoDescender")

    def test_vertTypoLineGap(self):
        self.checkInteger("openTypeVheaVertTypoLineGap")

    def test_vheaCaretSlopeRise(self):
        self.checkInteger("openTypeVheaCaretSlopeRise")

    def test_vheaCaretSlopeRun(self):
        self.checkInteger("openTypeVheaCaretSlopeRun")

    def test_vheaCaretOffset(self):
        self.checkInteger("openTypeVheaCaretOffset")

    # TODO: test_achVendorID()

    # TODO: test_usWidthClass(), test_usWeightClass()

    # TODO: test_fsSelection()

    # TODO: fsType

    # TODO: Panose

    # TODO: ulUnicodeRange

    def test_ulCodePageRange(self):
        codePageRangesEdit = self.tab.codePageRangesEdit
        codePageRangesEdit.setEnabled(True)
        model = codePageRangesEdit.model()
        for index in range(model.rowCount()):
            model.item(index).setCheckState(Qt.Checked)
        self.fontInfo.accept()

        codePageRanges = self.font.info.openTypeOS2CodePageRanges
        self.assertEqual(
            codePageRanges,
            [
                0,
                1,
                2,
                3,
                4,
                5,
                6,
                7,
                8,
                16,
                17,
                18,
                19,
                20,
                21,
                29,
                30,
                31,
                48,
                49,
                50,
                51,
                52,
                53,
                54,
                55,
                56,
                57,
                58,
                59,
                60,
                61,
                62,
                63,
            ],
        )

        codePageRangesEdit.setEnabled(False)
        self.fontInfo.accept()

        self.assertIsNone(self.font.info.openTypeOS2CodePageRanges)

    def test_sTypoAscender(self):
        self.checkInteger("openTypeOS2TypoAscender")

    def test_sTypoDescender(self):
        self.checkInteger("openTypeOS2TypoDescender")

    def test_sTypoLineGap(self):
        self.checkInteger("openTypeOS2TypoLineGap")

    def test_usWinAscent(self):
        self.checkPositiveInteger("openTypeOS2WinAscent")

    def test_usWinDescent(self):
        self.checkPositiveInteger("openTypeOS2WinDescent")

    def test_ySubscriptXSize(self):
        self.checkInteger("openTypeOS2SubscriptXSize")

    def test_ySubscriptYSize(self):
        self.checkInteger("openTypeOS2SubscriptYSize")

    def test_ySubscriptXOffset(self):
        self.checkInteger("openTypeOS2SubscriptXOffset")

    def test_ySubscriptYOffset(self):
        self.checkInteger("openTypeOS2SubscriptYOffset")

    def test_ySuperscriptXSize(self):
        self.checkInteger("openTypeOS2SuperscriptXSize")

    def test_ySuperscriptYSize(self):
        self.checkInteger("openTypeOS2SuperscriptYSize")

    def test_ySuperscriptXOffset(self):
        self.checkInteger("openTypeOS2SuperscriptXOffset")

    def test_ySuperscriptYOffset(self):
        self.checkInteger("openTypeOS2SuperscriptYOffset")

    def test_yStrikeoutSize(self):
        self.checkInteger("openTypeOS2StrikeoutSize")

    def test_yStrikeoutPosition(self):
        self.checkInteger("openTypeOS2StrikeoutPosition")


class PostScriptTabTest(TabTestCase):
    def __init__(self, methodName):
        super().__init__(methodName)

    def setUp(self):
        super().setUp()
        self.tab = self.postScriptTab

    def test_PostScriptTab_name(self):
        self.assertEqual(self.tab.name, "PostScript")

    def test_fontName(self):
        self.checkString("postscriptFontName")

    def test_fullName(self):
        self.checkString("postscriptFullName")

    def test_weightName(self):
        self.checkString("postscriptWeightName")

    def test_uniqueID(self):
        # XXX: Range?
        self.checkInteger("postscriptUniqueID")

    def test_blueValues(self):
        self.checkIntegerFloatList("postscriptBlueValues")

    def test_otherBlues(self):
        self.checkIntegerFloatList("postscriptOtherBlues")

    def test_familyBlues(self):
        self.checkIntegerFloatList("postscriptFamilyBlues")

    def test_familyOtherBlues(self):
        self.checkIntegerFloatList("postscriptFamilyOtherBlues")

    def test_stemSnapH(self):
        self.checkIntegerFloatList("postscriptStemSnapH")

    def test_stemSnapV(self):
        self.checkIntegerFloatList("postscriptStemSnapV")

    def test_blueFuzz(self):
        self.checkIntegerFloat("postscriptBlueFuzz")

    def test_blueScale(self):
        self.checkIntegerFloat("postscriptBlueScale")

    def test_blueShift(self):
        self.checkIntegerFloat("postscriptBlueShift")

    def test_forceBold(self):
        self.checkBoolean("postscriptForceBold")

    def test_defaultWidthX(self):
        self.checkIntegerFloat("postscriptDefaultWidthX")

    def test_nominalWidthX(self):
        self.checkIntegerFloat("postscriptNominalWidthX")

    def test_underlineThickness(self):
        self.checkIntegerFloat("postscriptUnderlineThickness")

    def test_underlinePosition(self):
        self.checkIntegerFloat("postscriptUnderlinePosition")

    def test_slantAngle(self):
        self.checkIntegerFloat("postscriptSlantAngle")

    def test_isFixedPitch(self):
        self.checkBoolean("postscriptIsFixedPitch")

    def test_defaultCharacter(self):
        self.checkString("postscriptDefaultCharacter")

    # TODO: test_windowsCharacterSet


if __name__ == "__main__":
    unittest.main()



"""
The *drawing* submodule
--------------------------------

The *drawing* submodule provides common drawing functions for glyph views.
It was adapted from defconAppKit and has similar APIs.

Notes:

- All drawing is done in font units
- The *scale* argument is the factor to scale a glyph unit to a view unit

"""

import math

from defcon import Color
from fontTools.misc.transform import Identity
from fontTools.pens.qtPen import QtPen
from fontTools.pens.transformPen import TransformPen
from PyQt5.QtCore import QLineF, QPointF, Qt
from PyQt5.QtGui import (
    QBrush,
    QColor,
    QPainter,
    QPainterPath,
    QPen,
    QPixmap,
    QTransform,
)
from PyQt5.QtWidgets import QApplication, QGraphicsPixmapItem, QGraphicsScene

# ------
# Colors
# ------

_defaultColors = dict(
    # General
    # -------
    background=QColor(Qt.white),
    # Font
    # ----
    # vertical metrics
    fontVerticalMetrics=QColor(204, 206, 200),
    fontPostscriptBlues=QColor(236, 209, 215, 100),
    fontPostscriptFamilyBlues=QColor.fromRgbF(1, 1, 0.5, 0.3),  # TODO: change that
    # guidelines
    fontGuideline=QColor.fromRgbF(1, 0, 0, 0.5),
    # Glyph
    # -----
    # contour fill
    glyphContourFill=QColor.fromRgbF(0.95, 0.95, 0.95, 0.3),
    # contour stroke
    glyphContourStroke=QColor(34, 34, 34),
    # component fill
    glyphComponentFill=QColor(90, 90, 90, 135),
    # component stroke
    glyphComponentStroke=QColor.fromRgbF(0, 0, 0, 1),
    # points
    glyphOnCurvePoints=QColor(4, 100, 166, 190),
    glyphOnCurveSmoothPoints=QColor(41, 172, 118, 190),
    glyphOffCurvePoints=QColor(116, 116, 116),
    glyphOtherPoints=QColor(140, 140, 140, 240),
    # anchors
    glyphAnchor=QColor(178, 102, 76, 200),
    # guidelines
    glyphGuideline=QColor.fromRgbF(0.3, 0.4, 0.85, 0.5),
    # marker
    glyphBluesMarker=QColor(235, 191, 202, 225),
)


def applyEffectToPixmap(pixmap, effect):
    scene = QGraphicsScene()
    item = QGraphicsPixmapItem()
    item.setPixmap(pixmap)
    item.setGraphicsEffect(effect)
    scene.addItem(item)
    res = QPixmap(pixmap.size())
    res.fill(Qt.transparent)
    painter = QPainter(res)
    scene.render(painter)
    return res


def colorToQColor(color):
    """
    Returns the QColor_ that corresponds to the defcon Color_ *color*.

    TODO: Color lacks online documentation.

    .. _Color: https://github.com/typesupply/defcon/blob/ufo3/Lib/defcon/objects/color.py
    .. _QColor: http://doc.qt.io/qt-5/qcolor.html
    """
    r, g, b, a = Color(color)
    return QColor.fromRgbF(r, g, b, a)


def defaultColor(name):
    """
    Returns a fallback QColor_ for a given *name*.

    TODO: name list?

    .. _QColor: http://doc.qt.io/qt-5/qcolor.html
    """
    return _defaultColors[name]


def ellipsePath(x, y, size):
    halfSize = size / 2
    path = QPainterPath()
    path.addEllipse(x - halfSize, y - halfSize, size, size)
    return path


def lozengePath(x, y, size):
    halfSize = size / 2
    path = QPainterPath()
    path.moveTo(x - halfSize, y)
    path.lineTo(x, y + halfSize)
    path.lineTo(x + halfSize, y)
    path.lineTo(x, y - halfSize)
    path.closeSubpath()
    return path


def rectanglePath(x, y, size):
    halfSize = size / 2
    path = QPainterPath()
    path.addRect(x - halfSize, y - halfSize, size, size)
    return path


def trianglePath(x, y, size, angle):
    thirdSize = size / 3
    pen = QtPen({})
    tPen = TransformPen(pen, Identity.rotate(angle))
    tPen.moveTo((-thirdSize, size))
    tPen.lineTo((-thirdSize, -size))
    tPen.lineTo((2 * thirdSize, 0))
    tPen.closePath()
    return pen.path.translated(x, y)


# ----------
# Primitives
# ----------


def drawLine(painter, x1, y1, x2, y2, lineWidth=0):
    """
    Draws a line from *(x1, y1)* to *(x2, y2)* with a thickness of *lineWidth*
    and using QPainter_ *painter*.

    Compared to the built-in ``painter.drawLine(…)`` method, this will disable
    antialiasing for horizontal/vertical lines.

    .. _`cosmetic pen`: http://doc.qt.io/qt-5/qpen.html#isCosmetic
    .. _QPainter: http://doc.qt.io/qt-5/qpainter.html
    """
    painter.save()
    if x1 == x2 or y1 == y2:
        painter.setRenderHint(QPainter.Antialiasing, False)
    pen = painter.pen()
    pen.setWidthF(lineWidth)
    painter.setPen(pen)
    painter.drawLine(QPointF(x1, y1), QPointF(x2, y2))
    painter.restore()


def drawTextAtPoint(
    painter, text, x, y, scale, xAlign="left", yAlign="bottom", flipped=True
):
    """
    Draws *text* at *(x, y)* with scale *scale* and a given QPainter_.

    - *xAlign* may be "left", "center" or "right" and specifies the alignment
      of the text (left-aligned text is painted to the right of *x*)
    - *yAlign* may be "top", "center" or "bottom" and specifies the y-positing
      of the text block relative to *y*

    TODO: support LTR http://stackoverflow.com/a/24831796/2037879

    .. _QPainter: http://doc.qt.io/qt-5/qpainter.html
    """
    fM = painter.fontMetrics()
    lines = text.splitlines()
    lineSpacing = fM.lineSpacing()
    if xAlign != "left" or yAlign != "bottom":
        width = scale * max(fM.width(line) for line in lines)
        height = scale * len(lines) * lineSpacing
        if xAlign == "center":
            x -= width / 2
        elif xAlign == "right":
            x -= width
        if yAlign == "center":
            y += height / 2
        elif yAlign == "top":
            y += height
    painter.save()
    if flipped:
        s = -scale
        height = fM.ascent() * scale
        y -= height
    else:
        s = scale
    painter.translate(x, y)
    painter.scale(scale, s)
    for line in lines:
        painter.drawText(0, 0, line)
        painter.translate(0, lineSpacing)
    painter.restore()


def drawTiles(painter, rect, tileSize=6, color=None, backgroundColor=None):
    sz = 2 * tileSize
    tiledPixmap = QPixmap(sz, sz)
    pixmapPainter = QPainter(tiledPixmap)
    pixmapPainter.setPen(Qt.NoPen)
    pixmapPainter.setBrush(Qt.Dense4Pattern)
    brush = pixmapPainter.brush()
    brush.setColor(color)
    pixmapPainter.setBrush(brush)
    pixmapPainter.setBackground(QBrush(backgroundColor))
    pixmapPainter.setBackgroundMode(Qt.OpaqueMode)
    pixmapPainter.scale(tileSize, tileSize)
    pixmapPainter.drawRect(tiledPixmap.rect())
    pixmapPainter.end()
    painter.drawTiledPixmap(rect, tiledPixmap)


# ----
# Font
# ----

# Guidelines


def drawFontGuidelines(
    painter, glyph, scale, rect, drawLines=True, drawText=True, color=None
):
    """
    Draws the font guidelines of the Glyph_ *glyph* in the form of lines if
    *drawLines* is true and text if *drawText* is true using QPainter_
    *painter*.

    *rect* specifies the rectangle which the lines will be drawn in (usually,
    that of the glyph’s advance width).

    .. _Glyph: http://ts-defcon.readthedocs.org/en/ufo3/objects/glyph.html
    .. _QPainter: http://doc.qt.io/qt-5/qpainter.html
    """
    if not (drawLines or drawText):
        return
    font = glyph.font
    if font is None:
        return
    if color is None:
        color = defaultColor("fontGuideline")
    _drawGuidelines(
        painter,
        glyph,
        scale,
        rect,
        font.guidelines,
        drawLines=drawLines,
        drawText=drawText,
        color=color,
    )


def drawGlyphGuidelines(
    painter, glyph, scale, rect, drawLines=True, drawText=True, color=None
):
    if not (drawLines or drawText):
        return
    if color is None:
        color = defaultColor("glyphGuideline")
    _drawGuidelines(
        painter,
        glyph,
        scale,
        rect,
        glyph.guidelines,
        drawLines=drawLines,
        drawText=drawText,
        color=color,
    )


def _drawGuidelines(
    painter, glyph, scale, rect, guidelines, drawLines=True, drawText=True, color=None
):
    if not (drawLines or drawText):
        return
    xMin, yMin, width, height = rect
    xMax = xMin + width
    yMax = yMin + height
    for line in guidelines:
        color_ = color
        if color_ is None:
            if line.color:
                color_ = colorToQColor(line.color)
            else:
                color_ = defaultColor("glyphGuideline")
        painter.save()
        painter.setPen(color)
        line1 = None
        if None not in (line.x, line.y):
            if line.angle is not None:
                # make an infinite line that intersects *(line.x, line.y)*
                # 1. make horizontal line from *(line.x, line.y)* of length *diagonal*
                diagonal = math.sqrt(width ** 2 + height ** 2)
                line1 = QLineF(line.x, line.y, line.x + diagonal, line.y)
                # 2. set the angle
                # defcon guidelines are clockwise
                line1.setAngle(line.angle)
                # 3. reverse the line and set length to 2 * *diagonal*
                line1.setPoints(line1.p2(), line1.p1())
                line1.setLength(2 * diagonal)
            else:
                line1 = QLineF(xMin, line.y, xMax, line.y)
        textX = 0
        textY = 0
        if drawLines:
            if line1 is not None:
                # line
                drawLine(painter, line1.x1(), line1.y1(), line1.x2(), line1.y2())
                # point
                x, y = line.x, line.y
                smoothWidth = 8 * scale
                smoothHalf = smoothWidth / 2.0
                painter.save()
                pointPath = QPainterPath()
                x -= smoothHalf
                y -= smoothHalf
                pointPath.addEllipse(x, y, smoothWidth, smoothWidth)
                pen = QPen(color_)
                pen.setWidthF(1 * scale)
                painter.setPen(pen)
                painter.drawPath(pointPath)
                painter.restore()
            else:
                if line.y is not None:
                    drawLine(painter, xMin, line.y, xMax, line.y)
                elif line.x is not None:
                    drawLine(painter, line.x, yMin, line.x, yMax)
        if drawText and line.name:
            if line1 is not None:
                textX = line.x
                textY = line.y - 6 * scale
                xAlign = "center"
            else:
                if line.y is not None:
                    fontSize = painter.font().pointSize()
                    textX = glyph.width + 6 * scale
                    textY = line.y - (fontSize / 3.5) * scale
                elif line.x is not None:
                    textX = line.x + 6 * scale
                    textY = 0
                xAlign = "left"
            drawTextAtPoint(painter, line.name, textX, textY, scale, xAlign=xAlign)
        painter.restore()


# Blues


def drawFontPostscriptBlues(painter, glyph, scale, color=None):
    """
    Draws a Glyph_ *glyph*’s blue values.

    .. _Glyph: http://ts-defcon.readthedocs.org/en/ufo3/objects/glyph.html
    """
    font = glyph.font
    if font is None:
        return
    blues = []
    if font.info.postscriptBlueValues:
        blues += font.info.postscriptBlueValues
    if font.info.postscriptOtherBlues:
        blues += font.info.postscriptOtherBlues
    if not blues:
        return
    if color is None:
        color = defaultColor("fontPostscriptBlues")
    _drawBlues(painter, glyph, blues, color)


def drawFontPostscriptFamilyBlues(painter, glyph, scale, color=None):
    """
    Draws a Glyph_ *glyph*’s family blue values.

    .. _Glyph: http://ts-defcon.readthedocs.org/en/ufo3/objects/glyph.html
    """
    font = glyph.font
    if font is None:
        return
    blues = []
    if font.info.postscriptFamilyBlues:
        blues += font.info.postscriptFamilyBlues
    if font.info.postscriptFamilyOtherBlues:
        blues += font.info.postscriptFamilyOtherBlues
    if not blues:
        return
    if color is None:
        color = defaultColor("fontPostscriptFamilyBlues")
    _drawBlues(painter, glyph, blues, color)


def _drawBlues(painter, glyph, blues, color):
    for yMin, yMax in zip(blues[::2], blues[1::2]):
        painter.fillRect(0, yMin, glyph.width, yMax - yMin, color)


# Image


def drawGlyphImage(painter, glyph, scale):
    """
    Draws a Glyph_ *glyph*’s image.

    .. _Glyph: http://ts-defcon.readthedocs.org/en/ufo3/objects/glyph.html
    """
    image = glyph.image
    pixmap = image.getRepresentation("defconQt.QPixmap")
    if pixmap is None:
        return
    painter.save()
    painter.setTransform(QTransform(*image.transformation), True)
    painter.translate(0, pixmap.height())
    painter.scale(1, -1)
    painter.drawPixmap(0, 0, pixmap)
    painter.restore()


# Metrics


def drawGlyphMetrics(
    painter,
    glyph,
    scale,
    drawHMetrics=True,
    drawVMetrics=True,
    drawText=False,
    color=None,
    textColor=None,
):
    """
    TODO: doc comment needs update

    Draws vertical metrics of the Glyph_ *glyph* (ascender, descender,
    baseline, x-height, cap height) in the form of lines if *drawLines* is true
    and text if *drawText* is true using QPainter_ *painter*.

    .. _Glyph: http://ts-defcon.readthedocs.org/en/ufo3/objects/glyph.html
    .. _QPainter: http://doc.qt.io/qt-5/qpainter.html
    """
    font = glyph.font
    if font is None:
        return
    if color is None:
        color = defaultColor("fontVerticalMetrics")
    if textColor is None:
        textColor = defaultColor("glyphOtherPoints")
    painter.save()
    pen = QPen(color)
    pen.setWidth(0)
    painter.setPen(pen)
    painter.setRenderHint(QPainter.Antialiasing, False)
    # metrics
    metrics = [
        font.info.descender or -250,
        0,
        font.info.xHeight,
        font.info.capHeight,
        font.info.ascender or 750,
    ]
    if drawHMetrics:
        lo = metrics[0]
        hi = max(y for y in metrics[-2:] if y is not None)
        painter.drawLine(0, lo, 0, hi)
        painter.drawLine(glyph.width, lo, glyph.width, hi)
    if drawVMetrics:
        for y in metrics:
            if y is None:
                continue
            painter.drawLine(0, y, glyph.width, y)
    painter.restore()
    # metrics text
    if drawText:
        toDraw = [
            (QApplication.translate("drawing", "Descender"), font.info.descender),
            (QApplication.translate("drawing", "Baseline"), 0),
            (QApplication.translate("drawing", "x-height"), font.info.xHeight),
            (QApplication.translate("drawing", "Cap height"), font.info.capHeight),
            (QApplication.translate("drawing", "Ascender"), font.info.ascender),
        ]
        # gather y positions
        positions = {}
        for name, position in toDraw:
            if position is None:
                continue
            if position not in positions:
                positions[position] = []
            positions[position].append(name)
        # create lines
        lines = []
        for y, names in positions.items():
            names = ", ".join(names)
            if y != 0:
                names = "%s (%d)" % (names, y)
            lines.append((y, names))
        painter.save()
        painter.setPen(textColor)
        fontSize = painter.font().pointSize()
        x = glyph.width + 6 * scale
        for y, text in lines:
            y -= (fontSize / 3.5) * scale
            drawTextAtPoint(painter, text, x, y, scale)
        painter.restore()


# Fill and Stroke


def drawGlyphFillAndStroke(
    painter,
    glyph,
    scale,
    drawFill=True,
    drawStroke=True,
    drawComponentFill=True,
    drawComponentStroke=False,
    contourFillColor=None,
    contourStrokeColor=None,
    componentFillColor=None,
    componentStrokeColor=None,
):
    """
    Draws a Glyph_ *glyph* contours’ fill and stroke.

    Component fill is always drawn, component stroke is drawn if
    *componentStrokeColor* is not None.

    .. _Glyph: http://ts-defcon.readthedocs.org/en/ufo3/objects/glyph.html
    """
    # get the layer color
    layer = glyph.layer
    layerColor = None
    if layer is not None and layer.color is not None:
        layerColor = colorToQColor(layer.color)
    # get the paths
    contourPath = glyph.getRepresentation("defconQt.NoComponentsQPainterPath")
    componentPath = glyph.getRepresentation("defconQt.OnlyComponentsQPainterPath")
    painter.save()
    # fill
    # contours
    if drawFill:
        if contourFillColor is None:
            if layerColor is not None:
                contourFillColor = layerColor
            else:
                contourFillColor = defaultColor("glyphContourFill")
        painter.fillPath(contourPath, QBrush(contourFillColor))
    # components
    if drawComponentFill:
        if componentFillColor is None:
            if layerColor is not None:
                componentFillColor = layerColor
            else:
                componentFillColor = defaultColor("glyphComponentFill")
        painter.fillPath(componentPath, QBrush(componentFillColor))
    # stroke
    if drawStroke:
        if contourStrokeColor is None:
            if layerColor is not None:
                contourStrokeColor = layerColor
            else:
                contourStrokeColor = defaultColor("glyphContourStroke")
        pen = QPen(contourStrokeColor)
        pen.setWidth(0)
        painter.setPen(pen)
        painter.drawPath(contourPath)
    # components
    if drawComponentStroke:
        if componentStrokeColor is None:
            if layerColor is not None:
                componentStrokeColor = layerColor
            else:
                componentStrokeColor = defaultColor("glyphContourStroke")
        pen = QPen(componentStrokeColor)
        pen.setWidth(0)
        painter.setPen(pen)
        painter.drawPath(componentPath)
    painter.restore()


# points


def drawGlyphPoints(
    painter,
    glyph,
    scale,
    drawStartPoints=True,
    drawOnCurves=True,
    drawOffCurves=True,
    drawCoordinates=False,
    drawBluesMarkers=True,
    onCurveColor=None,
    onCurveSmoothColor=None,
    offCurveColor=None,
    otherColor=None,
    backgroundColor=None,
):
    """
    Draws a Glyph_ *glyph*’s points.

    .. _Glyph: http://ts-defcon.readthedocs.org/en/ufo3/objects/glyph.html
    """
    if onCurveColor is None:
        onCurveColor = defaultColor("glyphOnCurvePoints")
    if onCurveSmoothColor is None:
        onCurveSmoothColor = defaultColor("glyphOnCurveSmoothPoints")
    if offCurveColor is None:
        offCurveColor = defaultColor("glyphOffCurvePoints")
    if otherColor is None:
        otherColor = defaultColor("glyphOtherPoints")
    if backgroundColor is None:
        backgroundColor = defaultColor("background")
    bluesMarkerColor = defaultColor("glyphBluesMarker")
    notchColor = defaultColor("glyphContourStroke").lighter(200)
    # get the outline data
    outlineData = glyph.getRepresentation("defconQt.OutlineInformation")
    points = []
    # blue zones markers
    if drawBluesMarkers and drawOnCurves:
        font = glyph.font
        blues = []
        if font.info.postscriptBlueValues:
            blues += font.info.postscriptBlueValues
        if font.info.postscriptOtherBlues:
            blues += font.info.postscriptOtherBlues
        if blues:
            blues_ = set(blues)
            size = 13 * scale
            snapSize = 17 * scale
            painter.save()
            pen = painter.pen()
            pen.setColor(QColor(255, 255, 255, 125))
            pen.setWidth(0)
            painter.setPen(pen)
            for point in outlineData["onCurvePoints"]:
                x, y = point["point"]
                # TODO: we could add a non-overlapping interval tree special
                # cased for borders
                for yMin, yMax in zip(blues[::2], blues[1::2]):
                    if not (y >= yMin and y <= yMax):
                        continue
                    # if yMin > 0 and y == yMin or yMin <= 0 and y == yMax:
                    if y in blues_:
                        path = lozengePath(x, y, snapSize)
                    else:
                        path = ellipsePath(x, y, size)
                    painter.fillPath(path, bluesMarkerColor)
                    painter.drawPath(path)
            painter.restore()
    # handles
    if drawOffCurves and outlineData["offCurvePoints"]:
        painter.save()
        painter.setPen(otherColor)
        for x1, y1, x2, y2 in outlineData["bezierHandles"]:
            drawLine(painter, x1, y1, x2, y2)
        painter.restore()
    # on curve
    if drawOnCurves and outlineData["onCurvePoints"]:
        size = 6.5 * scale
        smoothSize = 8 * scale
        startSize = 7 * scale
        loneStartSize = 12 * scale
        painter.save()
        notchPath = QPainterPath()
        path = QPainterPath()
        smoothPath = QPainterPath()
        for point in outlineData["onCurvePoints"]:
            x, y = point["point"]
            points.append((x, y))
            # notch
            if "smoothAngle" in point:
                angle = point["smoothAngle"]
                t = Identity.rotate(angle)
                x1, y1 = t.transformPoint((-1.35 * scale, 0))
                x2, y2 = -x1, -y1
                notchPath.moveTo(x1 + x, y1 + y)
                notchPath.lineTo(x2 + x, y2 + y)
            # points
            if drawStartPoints and "startPointAngle" in point:
                angle = point["startPointAngle"]
                if angle is not None:
                    pointPath = trianglePath(x, y, startSize, angle)
                else:
                    pointPath = ellipsePath(x, y, loneStartSize)
            elif point["smooth"]:
                pointPath = ellipsePath(x, y, smoothSize)
            else:
                pointPath = rectanglePath(x, y, size)
            # store the path
            if point["smooth"]:
                smoothPath.addPath(pointPath)
            else:
                path.addPath(pointPath)
        # stroke
        pen = QPen(onCurveColor)
        pen.setWidthF(1.2 * scale)
        painter.setPen(pen)
        painter.drawPath(path)
        pen.setColor(onCurveSmoothColor)
        painter.setPen(pen)
        painter.drawPath(smoothPath)
        # notch
        pen.setColor(notchColor)
        pen.setWidth(0)
        painter.setPen(pen)
        painter.drawPath(notchPath)
        painter.restore()
    # off curve
    if drawOffCurves and outlineData["offCurvePoints"]:
        # points
        offSize = 4.25 * scale
        path = QPainterPath()
        for point in outlineData["offCurvePoints"]:
            x, y = point["point"]
            pointPath = ellipsePath(x, y, offSize)
            path.addPath(pointPath)
        pen = QPen(offCurveColor)
        pen.setWidthF(2.5 * scale)
        painter.save()
        painter.setPen(pen)
        painter.drawPath(path)
        painter.fillPath(path, QBrush(backgroundColor))
        painter.restore()
    # coordinates
    if drawCoordinates:
        painter.save()
        painter.setPen(otherColor)
        font = painter.font()
        font.setPointSize(7)
        painter.setFont(font)
        for x, y in points:
            posX = x
            # TODO: We use + here because we align on top. Consider abstracting
            # yOffset.
            posY = y + 6 * scale
            x = round(x, 1)
            if int(x) == x:
                x = int(x)
            y = round(y, 1)
            if int(y) == y:
                y = int(y)
            text = "%d  %d" % (x, y)
            drawTextAtPoint(
                painter, text, posX, posY, scale, xAlign="center", yAlign="top"
            )
        painter.restore()


# Anchors


def drawGlyphAnchors(
    painter, glyph, scale, drawAnchors=True, drawText=True, color=None
):
    """
    Draws a Glyph_ *glyph*’s anchors.

    .. _Glyph: http://ts-defcon.readthedocs.org/en/ufo3/objects/glyph.html
    """
    if not glyph.anchors:
        return
    if color is None:
        color = defaultColor("glyphAnchor")
    fallbackColor = color
    anchorSize = 9 * scale
    for anchor in glyph.anchors:
        if anchor.color is not None:
            color = colorToQColor(anchor.color)
        else:
            color = fallbackColor
        x, y = anchor.x, anchor.y
        name = anchor.name
        painter.save()
        if drawAnchors:
            path = lozengePath(x, y, anchorSize)
            painter.fillPath(path, color)
        if drawText and name:
            painter.setPen(color)
            # TODO: we're using + before we shift to top, ideally this should
            # be abstracted w drawTextAtPoint taking a dy parameter that will
            # offset the drawing region from origin regardless of whether we
            # are aligning to top or bottom.
            y += 6 * scale
            drawTextAtPoint(painter, name, x, y, scale, xAlign="center", yAlign="top")
        painter.restore()



from PyQt5.QtCore import QSize, Qt
from PyQt5.QtGui import QColor, QPainter
from PyQt5.QtWidgets import QWidget


class GlyphStackWidget(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self._alignment = "left"
        self._buffer = 15
        self._glyphs = []
        self._maxWidth = 300

        self._backgroundColor = Qt.transparent
        self._glyphColor = QColor.fromRgbF(0, 0, 0, 0.15)

        self._upm = 1000
        self._descender = -250

    # -------------
    # Notifications
    # -------------

    def _subscribeToGlyphs(self, glyphs):
        handledGlyphs = set()
        handledFonts = set()
        for glyph in glyphs:
            if glyph in handledGlyphs:
                continue
            handledGlyphs.add(glyph)
            glyph.addObserver(self, "_glyphChanged", "Glyph.Changed")
            font = glyph.font
            if font is None:
                continue
            if font in handledFonts:
                continue
            handledFonts.add(font)
            font.info.addObserver(self, "_fontChanged", "Info.Changed")

    def _unsubscribeFromGlyphs(self):
        handledGlyphs = set()
        handledFonts = set()
        for glyph in self._glyphs:
            if glyph in handledGlyphs:
                continue
            handledGlyphs.add(glyph)
            glyph.removeObserver(self, "Glyph.Changed")
            font = glyph.font
            if font is None:
                continue
            if font in handledFonts:
                continue
            handledFonts.add(font)
            font.info.removeObserver(self, "Info.Changed")

    def _glyphChanged(self, notification):
        self.update()

    def _fontChanged(self, notification):
        self.setGlyphs(self._glyphs)

    # --------------
    # Custom methods
    # --------------

    def glyphs(self):
        return self._glyphs

    def setGlyphs(self, glyphs):
        self._unsubscribeFromGlyphs()
        self._glyphs = glyphs
        upms = []
        descenders = []
        for glyph in self._glyphs:
            font = glyph.font
            if font is not None:
                upm = font.info.unitsPerEm
                if upm is not None:
                    upms.append(upm)
                descender = font.info.descender
                if descender is not None:
                    descenders.append(descender)
        if upms:
            self._upm = max(upms)
        if descenders:
            self._descender = min(descenders)
        if self._glyphs:
            self._maxWidth = max(glyph.width for glyph in self._glyphs)
        else:
            self._maxWidth = 300
        self._subscribeToGlyphs(glyphs)
        self.update()

    def alignment(self):
        return self._alignment

    def setAlignment(self, alignment):
        assert alignment in ("left", "center", "right")
        self._alignment = alignment
        self.update()

    def backgroundColor(self):
        return self._backgroundColor

    def setBackgroundColor(self, color):
        self._backgroundColor = color

    def glyphColor(self):
        return self._glyphColor

    def setGlyphColor(self, color):
        self._glyphColor = color

    # --------
    # Qt methods
    # --------

    def sizeHint(self):
        return QSize(330, 400)

    def paintEvent(self, event):
        rect = event.rect()
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)

        painter.fillRect(rect, self._backgroundColor)
        if not self._glyphs:
            return

        painter.translate(0, self.height())
        painter.scale(1, -1)

        availableHeight = self.height() - 2 * self._buffer
        availableWidth = self.width() - 2 * self._buffer
        scale = availableHeight / (self._upm * 1.2)
        yOffset = abs(self._descender * scale) + self._buffer
        scale = min(scale, availableWidth / self._maxWidth)

        for glyph in self._glyphs:
            if self._alignment == "left":
                xOffset = self._buffer
            elif self._alignment == "center":
                xOffset = (self.width() - (glyph.width * scale)) / 2
            else:
                xOffset = availableWidth + self._buffer - glyph.width * scale
            path = glyph.getRepresentation("defconQt.QPainterPath")
            painter.save()
            painter.translate(xOffset, yOffset)
            painter.scale(scale, scale)
            painter.fillPath(path, self._glyphColor)
            painter.restore()



from PyQt5.QtCore import QObject, Qt
from PyQt5.QtGui import QColor, QCursor, QPainter, QPainterPath, QPixmap
from PyQt5.QtWidgets import QApplication, QGraphicsDropShadowEffect

from defconQt.tools.drawing import applyEffectToPixmap

_path = QPainterPath()
_path.moveTo(9, 7.3)
_path.lineTo(9, 24)
_path.lineTo(21, 12)
_path.lineTo(16.3, 12)
_path.lineTo(18.6, 6.6)
_path.lineTo(14.85, 5)
_path.lineTo(12.5, 10.7)
_path.closeSubpath()

path = QPainterPath()
path.moveTo(10, 9.75)
path.lineTo(12.8, 12.5)
path.lineTo(15.3, 6.5)
path.lineTo(17.2, 7.3)
path.lineTo(14.75, 13.1)
path.lineTo(18.5, 13.1)
path.lineTo(10, 21.5)
path.closeSubpath()


class BaseTool(QObject):
    icon = QPainterPath()
    name = QApplication.translate("BaseTool", "Tool")
    shortcut = None
    grabKeyboard = False

    @property
    def cursor(self):
        # TODO: cache?
        return self.makeCursor(_path, path, 9.5, 1)

    def toolActivated(self):
        pass

    def toolDisabled(self):
        pass

    def drawingAttribute(self, attr, flags):
        return None

    def drawingColor(self, attr, flags):
        return None

    @property
    def _font(self):
        return self.parent().window().font_()

    @property
    def _glyph(self):
        return self.parent().activeGlyph()

    # helper functions

    def clampToOrigin(self, pos, origin):
        deltaX = pos.x() - origin.x()
        deltaY = pos.y() - origin.y()
        # go into the first quadrant to simplify our study
        aDeltaX = abs(deltaX)
        aDeltaY = abs(deltaY)
        # diagonal incr.
        # if aDeltaY >= aDeltaX * 2:
        #     pos.setX(origin.x())
        # elif aDeltaY > aDeltaX / 2:
        #     avg = (aDeltaX + aDeltaY) / 2
        #     pos.setX(origin.x() + copysign(avg, deltaX))
        #     pos.setY(origin.y() + copysign(avg, deltaY))
        if aDeltaY >= aDeltaX:
            pos.setX(origin.x())
        else:
            pos.setY(origin.y())
        return pos

    def magnetPos(self, pos):
        widget = self.parent()
        mouseItem = widget.itemAt(pos)
        if isinstance(mouseItem, tuple):
            contour, index = mouseItem
            point = contour[index]
            pos.setX(point.x)
            pos.setY(point.y)
        # TODO: also clamp to (0, 0) and (glyph.width, 0), conditionally?
        return pos

    def makeCursor(self, whitePath, blackPath, x, y):
        pixmap = QPixmap(24, 24)
        pixmap.fill(Qt.transparent)
        painter = QPainter()
        painter.begin(pixmap)
        painter.setRenderHint(QPainter.Antialiasing)
        painter.translate(0, pixmap.height())
        painter.scale(1, -1)
        painter.fillPath(whitePath, Qt.white)
        painter.end()
        effect = QGraphicsDropShadowEffect()
        effect.setColor(QColor.fromRgbF(0, 0, 0, 0.3))
        effect.setBlurRadius(4)
        effect.setOffset(0, 1)
        pixmap = applyEffectToPixmap(pixmap, effect)
        painter.begin(pixmap)
        painter.setRenderHint(QPainter.Antialiasing)
        painter.translate(0, pixmap.height())
        painter.scale(1, -1)
        painter.fillPath(blackPath, Qt.black)
        painter.end()
        return QCursor(pixmap, x, y)

    # events

    def contextMenuEvent(self, event):
        pass

    def keyPressEvent(self, event):
        pass

    def keyReleaseEvent(self, event):
        pass

    def mousePressEvent(self, event):
        if event.button() == Qt.MidButton:
            self._panOrigin = event.globalPos()

    def mouseMoveEvent(self, event):
        if hasattr(self, "_panOrigin"):
            pos = event.globalPos()
            self.parent().scrollBy(pos - self._panOrigin)
            self._panOrigin = pos

    def mouseReleaseEvent(self, event):
        if hasattr(self, "_panOrigin"):
            del self._panOrigin

    def mouseDoubleClickEvent(self, event):
        pass

    # custom painting

    def paintBackground(self, painter, index):
        pass

    def paint(self, painter, index):
        pass



"""
Windows that want to plug-in their own menu entries must implement

- setupMenu(menuBar)
- menuBar()
- setMenuBar(menuBar)

"""
from PyQt5.QtGui import QKeySequence
from PyQt5.QtWidgets import QAction, QApplication, QMenu, QMenuBar

from trufont.tools import platformSpecific

MAX_RECENT_FILES = 10


class MenuBar(QMenuBar):
    def __init__(self, parent=None):
        super().__init__(parent)
        self._spawnElementsHint = True

    def shouldSpawnElements(self):
        return self.parent() is None or self._spawnElementsHint

    def spawnElementsHint(self):
        return self._spawnElementsHint

    def setSpawnElementsHint(self, value):
        self._spawnElementsHint = value

    def fetchMenu(self, title):
        title = _trMenuString(title)
        # cache lookup
        child = None
        for child_ in self.children():
            if not isinstance(child_, QMenu):
                continue
            if child_.title() == title:
                child = child_
        if child is not None:
            return child
        # spawn
        menu = Menu(title, self)
        if self.shouldSpawnElements():
            self.addMenu(menu)
        return menu

    def resetState(self):
        for menu in self.children():
            if not isinstance(menu, QMenu):
                continue
            menu.resetState()


class Menu(QMenu):
    def shouldSpawnElements(self):
        parent = self.parent()
        if parent is not None:
            return parent.shouldSpawnElements()
        return False

    def fetchAction(self, text, callback=None, shortcut=None):
        if shortcut is None:
            shortcut = _shortcuts.get(text)
        text = _trMenuString(text)
        # cache lookup
        action = None
        for action_ in self.actions():
            if action_.text() == text:
                action = action_
        # spawn
        if action is None:
            action = QAction(text, self)
            if self.shouldSpawnElements():
                self.addAction(action)
        # connect
        action.setEnabled(True)
        try:
            action.triggered.disconnect()
        except TypeError:
            pass
        if callback is not None:
            action.triggered.connect(lambda: callback())
        action.setShortcut(QKeySequence(shortcut))
        return action

    fetchMenu = MenuBar.fetchMenu

    def resetState(self):
        self._ready = True
        # TODO: reset submenus too?
        for action in self.actions():
            action.setEnabled(False)


class Entries:
    File = "&File"
    File_New = "&New…"
    File_Open = "&Open…"
    File_Open_Recent = "Open &Recent"
    File_Import = "&Import…"
    File_Save = "&Save"
    File_Save_As = "Save &As…"
    File_Save_All = "Save A&ll"
    File_Close = "&Close"
    File_Reload = "&Revert"
    File_Export = "&Export…"
    File_Exit = "E&xit"

    Edit = "&Edit"
    Edit_Undo = "&Undo"
    Edit_Redo = "&Redo"
    Edit_Cut = "C&ut"
    Edit_Copy = "&Copy"
    Edit_Copy_As_Component = "Copy &As Component"
    Edit_Paste = "&Paste"
    Edit_Clear = "Cl&ear"
    Edit_Select_All = "&Select All"
    Edit_Deselect = "&Deselect"
    Edit_Find = "&Find…"
    Edit_Settings = "&Settings…"

    View = "&View"
    View_Zoom_In = "Zoom &In"
    View_Zoom_Out = "Zoom &Out"
    View_Reset_Zoom = "&Reset Zoom"
    View_Next_Tab = "&Next Tab"
    View_Previous_Tab = "&Previous Tab"
    View_Next_Glyph = "&Next Glyph"
    View_Previous_Glyph = "&Previous Glyph"
    View_Layer_Up = "Layer &Up"
    View_Layer_Down = "Layer &Down"
    View_Show_Points = "Show P&oints"
    View_Show_Coordinates = "Show &Coordinates"
    View_Show_Coordinates_When_Selected = "Show Coordinates When &Selected"
    View_Show_Point_Coordinates = "Always Show &Point Coordinates"
    View_Show_Bezier_Handles_Coordinates = "Always Show &Bezier Handles Coordinates"
    View_Show_Metrics = "Show &Metrics"
    View_Show_Images = "Show &Images"
    View_Show_Guidelines = "Show &Guidelines"

    Font = "F&ont"
    Font_Font_Info = "Font &Info"
    Font_Font_Features = "Font &Features"
    Font_Add_Glyphs = "&Add Glyphs…"
    Font_Sort = "&Sort…"

    Scripts = "&Scripts"
    Scripts_Build_Extension = "&Build Extension…"

    # TODO: remove metrics
    Window = "&Window"
    Window_Minimize = "&Minimize"
    Window_Minimize_All = "Minimize &All"
    Window_Zoom = "&Zoom"
    Window_Groups = "&Groups"
    Window_Kerning = "&Kerning"
    Window_Metrics = "M&etrics"
    Window_Scripting = "&Scripting"
    Window_Properties = "&Properties"
    Window_Output = "&Output"

    Help = "&Help"
    Help_Documentation = "&Documentation"
    Help_Report_An_Issue = "&Report an Issue"
    Help_About = "&About"


_shortcuts = {
    Entries.File_New: QKeySequence.New,
    Entries.File_Open: QKeySequence.Open,
    Entries.File_Save: QKeySequence.Save,
    Entries.File_Save_As: QKeySequence.SaveAs,
    Entries.File_Close: platformSpecific.closeKeySequence(),
    Entries.File_Export: "Ctrl+E",
    Entries.File_Exit: QKeySequence.Quit,
    Entries.Edit_Undo: QKeySequence.Undo,
    Entries.Edit_Redo: QKeySequence.Redo,
    Entries.Edit_Cut: QKeySequence.Cut,
    Entries.Edit_Copy: QKeySequence.Copy,
    Entries.Edit_Copy_As_Component: "Ctrl+Alt+C",
    Entries.Edit_Paste: QKeySequence.Paste,
    Entries.Edit_Select_All: QKeySequence.SelectAll,
    Entries.Edit_Deselect: "Ctrl+D",
    Entries.Edit_Find: QKeySequence.Find,
    Entries.View_Zoom_In: QKeySequence.ZoomIn,
    Entries.View_Zoom_Out: QKeySequence.ZoomOut,
    Entries.View_Reset_Zoom: "Ctrl+0",
    Entries.View_Next_Tab: QKeySequence.NextChild,
    Entries.View_Previous_Tab: platformSpecific.previousTabSequence(),
    Entries.View_Next_Glyph: "End",
    Entries.View_Previous_Glyph: "Home",
    Entries.View_Layer_Up: "PgUp",
    Entries.View_Layer_Down: "PgDown",
    Entries.View_Show_Points: "Ctrl+Shift+N",
    Entries.View_Show_Metrics: "Ctrl+Shift+M",
    Entries.View_Show_Guidelines: "Ctrl+Shift+G",
    Entries.Font_Font_Info: "Ctrl+Alt+I",
    Entries.Font_Font_Features: "Ctrl+Alt+F",
    Entries.Font_Add_Glyphs: "Ctrl+G",
    Entries.Window_Minimize: "Ctrl+M",
    Entries.Window_Groups: "Ctrl+Alt+G",
    Entries.Window_Kerning: "Ctrl+Alt+K",
    Entries.Window_Metrics: "Ctrl+Alt+S",
    Entries.Window_Scripting: "Ctrl+Alt+R",
    Entries.Window_Properties: "Ctrl+Alt+P",
    Entries.Window_Output: "Ctrl+Alt+O",
}


def globalMenuBar():
    menuBar = MenuBar()
    fileMenu = menuBar.fetchMenu(Entries.File)
    fileMenu.fetchAction(Entries.File_New)
    fileMenu.fetchAction(Entries.File_Open)
    fileMenu.fetchMenu(Entries.File_Open_Recent)
    # no-op, caller will maintain this
    if not platformSpecific.mergeOpenAndImport():
        fileMenu.fetchAction(Entries.File_Import)
    fileMenu.addSeparator()
    fileMenu.fetchAction(Entries.File_Save)
    fileMenu.fetchAction(Entries.File_Save_As)
    fileMenu.fetchAction(Entries.File_Save_All)
    fileMenu.fetchAction(Entries.File_Close)
    fileMenu.fetchAction(Entries.File_Reload)
    fileMenu.addSeparator()
    fileMenu.fetchAction(Entries.File_Export)
    fileMenu.fetchAction(Entries.File_Exit)

    editMenu = menuBar.fetchMenu(Entries.Edit)
    editMenu.fetchAction(Entries.Edit_Undo)
    editMenu.fetchAction(Entries.Edit_Redo)
    editMenu.addSeparator()
    editMenu.fetchAction(Entries.Edit_Cut)
    editMenu.fetchAction(Entries.Edit_Copy)
    editMenu.fetchAction(Entries.Edit_Copy_As_Component)
    editMenu.fetchAction(Entries.Edit_Paste)
    editMenu.fetchAction(Entries.Edit_Clear)
    editMenu.fetchAction(Entries.Edit_Select_All)
    # editMenu.fetchAction(Entries.Edit_Deselect)
    editMenu.fetchAction(Entries.Edit_Find)
    editMenu.addSeparator()
    editMenu.fetchAction(Entries.Edit_Settings)
    menuBar.addMenu(editMenu)

    viewMenu = menuBar.fetchMenu(Entries.View)
    viewMenu.fetchAction(Entries.View_Zoom_In)
    viewMenu.fetchAction(Entries.View_Zoom_Out)
    viewMenu.fetchAction(Entries.View_Reset_Zoom)
    viewMenu.addSeparator()
    viewMenu.fetchAction(Entries.View_Next_Tab)
    viewMenu.fetchAction(Entries.View_Previous_Tab)
    viewMenu.fetchAction(Entries.View_Next_Glyph)
    viewMenu.fetchAction(Entries.View_Previous_Glyph)
    viewMenu.fetchAction(Entries.View_Layer_Up)
    viewMenu.fetchAction(Entries.View_Layer_Down)
    viewMenu.addSeparator()
    viewMenu.fetchAction(Entries.View_Show_Points)
    coordinatesSubmenu = viewMenu.fetchMenu(Entries.View_Show_Coordinates)
    coordinatesSubmenu.fetchAction(Entries.View_Show_Coordinates_When_Selected)
    coordinatesSubmenu.fetchAction(Entries.View_Show_Point_Coordinates)
    coordinatesSubmenu.fetchAction(Entries.View_Show_Bezier_Handles_Coordinates)
    viewMenu.fetchAction(Entries.View_Show_Metrics)
    viewMenu.fetchAction(Entries.View_Show_Images)
    viewMenu.fetchAction(Entries.View_Show_Guidelines)

    fontMenu = menuBar.fetchMenu(Entries.Font)
    fontMenu.fetchAction(Entries.Font_Font_Info)
    fontMenu.fetchAction(Entries.Font_Font_Features)
    fontMenu.addSeparator()
    fontMenu.fetchAction(Entries.Font_Add_Glyphs)
    fontMenu.fetchAction(Entries.Font_Sort)

    menuBar.fetchMenu(Entries.Scripts)
    # no-op, caller will maintain this

    windowMenu = menuBar.fetchMenu(Entries.Window)
    if platformSpecific.windowCommandsInMenu():
        windowMenu.fetchAction(Entries.Window_Minimize)
        windowMenu.fetchAction(Entries.Window_Minimize_All)
        windowMenu.fetchAction(Entries.Window_Zoom)
        windowMenu.addSeparator()
    windowMenu.fetchAction(Entries.Window_Groups)
    windowMenu.fetchAction(Entries.Window_Kerning)
    windowMenu.fetchAction(Entries.Window_Metrics)
    windowMenu.fetchAction(Entries.Window_Scripting)
    windowMenu.fetchAction(Entries.Window_Properties)
    windowMenu.addSeparator()
    windowMenu.fetchAction(Entries.Window_Output)

    helpMenu = menuBar.fetchMenu(Entries.Help)
    helpMenu.fetchAction(Entries.Help_Documentation)
    helpMenu.fetchAction(Entries.Help_Report_An_Issue)
    helpMenu.addSeparator()
    helpMenu.fetchAction(Entries.Help_About)

    return menuBar


def _trMenuString(string):
    return QApplication.translate("AppMenu", string)



import sys
import traceback

from PyQt5.QtWidgets import QMessageBox

_showMessages = True


def showCriticalException(e, message=None):
    _prepareException(e, QMessageBox.Critical, message)


def showWarningException(e, message=None):
    _prepareException(e, QMessageBox.Warning, message)


def exceptionCallback(etype, value, tb):
    title = ":("
    message = "TruFont has encountered a problem and must shutdown."
    _displayException(etype, value, tb, QMessageBox.Critical, title, message)


def _displayException(etype, value, tb, kind, title, message):
    global _showMessages
    exc = traceback.format_exception(etype, value, tb)
    exc_text = "".join(exc)
    print(exc_text, file=sys.stderr)

    if _showMessages:
        messageBox = QMessageBox(kind, title, message)
        standardButtons = QMessageBox.Ok | QMessageBox.Close
        if kind == QMessageBox.Critical:
            standardButtons |= QMessageBox.Ignore
        messageBox.setStandardButtons(standardButtons)
        messageBox.setDetailedText(exc_text)
        messageBox.setInformativeText(str(value))
        result = messageBox.exec_()
        if result == QMessageBox.Close:
            sys.exit(1)
        elif result == QMessageBox.Ignore:
            _showMessages = False


def _prepareException(e, kind, message):
    title = e.__class__.__name__
    _displayException(e.__class__, e, e.__traceback__, kind, title, message)



"""
Icon database, issues drawing commands. PathButton is a consumer.
"""
from PyQt5.QtCore import QSize
from PyQt5.QtGui import QColor, QPainterPath, QTransform

from trufont.objects.pathIcon import PathIcon

_listIconsStrokeColor = QColor(90, 90, 90)
_sidebarIconsFillColor = QColor(187, 187, 187)
_sidebarIconsStrokeColor = QColor(103, 103, 103)


def p_minus():
    path = QPainterPath()
    path.moveTo(4, 8)
    path.lineTo(12, 8)
    return path


def dc_minus():
    return [QSize(16, 16), (p_minus(), "2", _listIconsStrokeColor)]


def dc_plus():
    path = p_minus()
    path.moveTo(8, 4)
    path.lineTo(8, 12)
    return [QSize(16, 16), (path, "2", _listIconsStrokeColor)]


def i_minus():
    icon = PathIcon(16, 16)
    path = p_minus()
    icon.addStrokePath(path, _listIconsStrokeColor, 2)
    return icon


def i_plus():
    icon = PathIcon(16, 16)
    path = p_minus()
    path.moveTo(8, 4)
    path.lineTo(8, 12)
    icon.addStrokePath(path, _listIconsStrokeColor, 2)
    return icon


def p_down():
    path = QPainterPath()
    path.moveTo(8, 4)
    path.lineTo(8, 12)
    path.moveTo(4, 8)
    path.lineTo(8, 12)
    path.lineTo(12, 8)
    return path


def dc_down():
    return [QSize(16, 16), (p_down(), "2a", _listIconsStrokeColor)]


def dc_up():
    path = p_down() * QTransform.fromScale(1, -1).translate(0, -16)
    return [QSize(16, 16), (path, "2a", _listIconsStrokeColor)]


def i_warning():
    icon = PathIcon(20, 20)
    path = QPainterPath()
    path.moveTo(4, 16)
    path.lineTo(16, 16)
    path.lineTo(10, 5)
    path.closeSubpath()
    path_ = QPainterPath()
    path_.addRect(9, 8, 2, 4)
    path_.addRect(9, 13, 2, 2)
    icon.addFillPath(path - path_, QColor(230, 20, 20), antialiasing=True)
    return icon


def p_ellipses():
    circ1 = QPainterPath()
    circ1.addEllipse(5, 1, 14, 14)
    circ2 = QPainterPath()
    circ2.addEllipse(1, 9, 10, 10)
    return circ1, circ2


def dc_invscale():
    circ1, circ2 = p_ellipses()
    path = circ1 - circ2
    path.addPath(circ2)
    return [
        QSize(20, 20),
        (circ2, "f", _sidebarIconsFillColor),
        (path, "1a", _sidebarIconsStrokeColor),
    ]


def dc_scale():
    circ1, circ2 = p_ellipses()
    path = circ2 - circ1
    path.addPath(circ1)
    return [
        QSize(20, 20),
        (circ1, "f", _sidebarIconsFillColor),
        (path, "1a", _sidebarIconsStrokeColor),
    ]


def p_rotate():
    path = QPainterPath()
    path.moveTo(18, 11)
    path.cubicTo(17, 15.5, 14, 18, 10, 18)
    path.cubicTo(5.5, 18, 2, 14.5, 2, 10)
    path.cubicTo(2, 6, 5.5, 2, 10, 2)
    path.cubicTo(14, 2, 17, 4, 17, 7)
    path.moveTo(18, 4)
    path.lineTo(18, 7)
    path.lineTo(15, 7)
    path.addRect(9, 9, 2, 2)
    return path


def dc_rotate():
    t = QTransform.fromScale(-1, 1).translate(-20, 0)
    return [QSize(20, 20), (p_rotate() * t, "1a", _sidebarIconsStrokeColor)]


def dc_invrotate():
    return [QSize(20, 20), (p_rotate(), "1a", _sidebarIconsStrokeColor)]


def p_skew():
    path1 = QPainterPath()
    path1.moveTo(11, 2)
    path1.lineTo(18, 2)
    path1.lineTo(10, 18)
    path1.lineTo(3, 18)
    path1.closeSubpath()
    path2 = QPainterPath(path1)
    path2.moveTo(8, 2)
    path2.lineTo(2, 2)
    path2.lineTo(2, 14)
    return path1, path2


def dc_invskew():
    path1, path2 = p_skew()
    t = QTransform.fromScale(-1, 1).translate(-20, 0)
    return [
        QSize(20, 20),
        (path1 * t, "f", _sidebarIconsFillColor),
        (path2 * t, "1", _sidebarIconsStrokeColor),
    ]


def dc_skew():
    path1, path2 = p_skew()
    return [
        QSize(20, 20),
        (path1, "f", _sidebarIconsFillColor),
        (path2, "1", _sidebarIconsStrokeColor),
    ]


def dc_snap():
    path = QPainterPath()
    path.addEllipse(8, 8, 5, 5)
    path_ = QPainterPath()
    path_.moveTo(10, 2)
    path_.lineTo(10, 18)
    path_.moveTo(2, 10)
    path_.lineTo(18, 10)
    return [
        QSize(20, 20),
        (path, "f", _sidebarIconsFillColor),
        (path_, "1", _sidebarIconsStrokeColor),
    ]


def p_cubes():
    notch = 5
    cube1 = QPainterPath()
    cube1.moveTo(2, 2)
    cube1.lineTo(18 - notch, 2)
    cube1.lineTo(18 - notch, 18 - notch)
    cube1.lineTo(2, 18 - notch)
    cube1.closeSubpath()
    cube2 = QPainterPath()
    cube2.moveTo(2 + notch, 2 + notch)
    cube2.lineTo(18, 2 + notch)
    cube2.lineTo(18, 18)
    cube2.lineTo(2 + notch, 18)
    cube2.closeSubpath()
    return cube1, cube2


def dc_union():
    cube1, cube2 = p_cubes()
    path = cube1 + cube2
    return [
        QSize(20, 20),
        (path, "f", _sidebarIconsFillColor),
        (path, "1", _sidebarIconsStrokeColor),
    ]


def dc_subtract():
    cube1, cube2 = p_cubes()
    path = cube1 - cube2
    path_ = cube2
    path_.addPath(path)
    return [
        QSize(20, 20),
        (path, "f", _sidebarIconsFillColor),
        (path_, "1", _sidebarIconsStrokeColor),
    ]


def dc_intersect():
    cube1, cube2 = p_cubes()
    intersect = cube1 & cube2
    cubes = cube1
    cubes.addPath(cube2)
    return [
        QSize(20, 20),
        (intersect, "f", _sidebarIconsFillColor),
        (cubes, "1", _sidebarIconsStrokeColor),
    ]


def dc_xor():
    cube1, cube2 = p_cubes()
    intersect = cube1 & cube2
    cubes = cube1
    cubes.addPath(cube2)
    return [
        QSize(20, 20),
        (cubes - intersect, "f", _sidebarIconsFillColor),
        (cubes, "1", _sidebarIconsStrokeColor),
    ]


def p_mirror():
    raise NotImplementedError


def dc_hmirror():
    path = QPainterPath()
    path.moveTo(8, 18)
    path.lineTo(1, 18)
    path.lineTo(8, 1)
    path.closeSubpath()
    path_ = QPainterPath(path)
    path_.moveTo(11, 18)
    path_.lineTo(19, 18)
    path_.lineTo(11, 1)
    path_.closeSubpath()
    return [
        QSize(20, 20),
        (path, "f", _sidebarIconsFillColor),
        (path_, "1", _sidebarIconsStrokeColor),
    ]


def dc_vmirror():
    path = QPainterPath()
    path.moveTo(2, 1)
    path.lineTo(19, 8)
    path.lineTo(2, 8)
    path.closeSubpath()
    path_ = QPainterPath(path)
    path_.moveTo(2, 11)
    path_.lineTo(19, 11)
    path_.lineTo(2, 19)
    path_.closeSubpath()
    return [
        QSize(20, 20),
        (path, "f", _sidebarIconsFillColor),
        (path_, "1", _sidebarIconsStrokeColor),
    ]


def dc_alignhleft():
    path = QPainterPath()
    path.moveTo(2, 2)
    path.lineTo(2, 18)
    path.lineTo(3, 18)
    path.lineTo(3, 2)
    path.closeSubpath()
    path.moveTo(3, 4)
    path.lineTo(13, 4)
    path.lineTo(13, 8)
    path.lineTo(3, 8)
    path.moveTo(3, 12)
    path.lineTo(18, 12)
    path.lineTo(18, 16)
    path.lineTo(3, 16)
    return [QSize(20, 20), (path, "1", _sidebarIconsStrokeColor)]


def dc_alignhcenter():
    path = QPainterPath()
    path.moveTo(9, 2)
    path.lineTo(9, 18)
    path.lineTo(10, 18)
    path.lineTo(10, 2)
    path.closeSubpath()
    path.moveTo(5, 4)
    path.lineTo(15, 4)
    path.lineTo(15, 8)
    path.lineTo(5, 8)
    path.closeSubpath()
    path.moveTo(2, 12)
    path.lineTo(18, 12)
    path.lineTo(18, 16)
    path.lineTo(2, 16)
    path.closeSubpath()
    return [QSize(20, 20), (path, "1", _sidebarIconsStrokeColor)]


def dc_alignhright():
    path = QPainterPath()
    path.moveTo(17, 2)
    path.lineTo(17, 18)
    path.lineTo(18, 18)
    path.lineTo(18, 2)
    path.closeSubpath()
    path.moveTo(17, 4)
    path.lineTo(8, 4)
    path.lineTo(8, 8)
    path.lineTo(17, 8)
    path.moveTo(17, 12)
    path.lineTo(2, 12)
    path.lineTo(2, 16)
    path.lineTo(17, 16)
    return [QSize(20, 20), (path, "1", _sidebarIconsStrokeColor)]


def dc_alignvtop():
    path = QPainterPath()
    path.moveTo(2, 2)
    path.lineTo(18, 2)
    path.lineTo(18, 3)
    path.lineTo(2, 3)
    path.closeSubpath()
    path.moveTo(4, 3)
    path.lineTo(4, 18)
    path.lineTo(8, 18)
    path.lineTo(8, 3)
    path.moveTo(12, 3)
    path.lineTo(12, 13)
    path.lineTo(16, 13)
    path.lineTo(16, 3)
    return [QSize(20, 20), (path, "1", _sidebarIconsStrokeColor)]


def dc_alignvcenter():
    path = QPainterPath()
    path.moveTo(2, 9)
    path.lineTo(18, 9)
    path.lineTo(18, 10)
    path.lineTo(2, 10)
    path.closeSubpath()
    path.moveTo(4, 2)
    path.lineTo(4, 18)
    path.lineTo(8, 18)
    path.lineTo(8, 2)
    path.closeSubpath()
    path.moveTo(12, 5)
    path.lineTo(12, 15)
    path.lineTo(16, 15)
    path.lineTo(16, 5)
    path.closeSubpath()
    return [QSize(20, 20), (path, "1", _sidebarIconsStrokeColor)]


def dc_alignvbottom():
    path = QPainterPath()
    path.moveTo(2, 17)
    path.lineTo(18, 17)
    path.lineTo(18, 18)
    path.lineTo(2, 18)
    path.closeSubpath()
    path.moveTo(4, 17)
    path.lineTo(4, 2)
    path.lineTo(8, 2)
    path.lineTo(8, 17)
    path.moveTo(12, 17)
    path.lineTo(12, 8)
    path.lineTo(16, 8)
    path.lineTo(16, 17)
    return [QSize(20, 20), (path, "1", _sidebarIconsStrokeColor)]



from PyQt5.QtCore import Qt
from PyQt5.QtGui import QTextCursor
from PyQt5.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QGridLayout,
    QGroupBox,
    QPlainTextEdit,
    QPushButton,
    QRadioButton,
    QVBoxLayout,
)

from defconQt.controls.colorVignette import ColorVignette
from trufont.objects import settings

sortItems = [
    "alphabetical",
    "category",
    "unicode",
    "script",
    "suffix",
    "decompositionBase",
    "weightedSuffix",
    "ligature",
]


class AddGlyphsDialog(QDialog):

    # TODO: implement Frederik's Glyph Construction Builder
    def __init__(self, currentGlyphs=None, parent=None):
        super().__init__(parent)
        self.setWindowModality(Qt.WindowModal)
        self.setWindowTitle(self.tr("Add Glyphs…"))
        self.currentGlyphs = currentGlyphs
        self.currentGlyphNames = [glyph.name for glyph in currentGlyphs]

        layout = QGridLayout(self)
        self.markColorWidget = ColorVignette(self)
        self.markColorWidget.setFixedWidth(56)
        self.importCharDrop = QComboBox(self)
        self.importCharDrop.addItem(self.tr("Import glyph names…"))
        glyphSets = settings.readGlyphSets()
        for name, glyphNames in glyphSets.items():
            self.importCharDrop.addItem(name, glyphNames)
        self.importCharDrop.currentIndexChanged[int].connect(self.importGlyphs)
        self.addGlyphsEdit = QPlainTextEdit(self)
        self.addGlyphsEdit.setFocus(Qt.OtherFocusReason)

        self.addUnicodeBox = QCheckBox(self.tr("Add Unicode"), self)
        self.addUnicodeBox.setChecked(True)
        self.addAsTemplateBox = QCheckBox(self.tr("Add as template"), self)
        self.addAsTemplateBox.setChecked(True)
        self.sortFontBox = QCheckBox(self.tr("Sort font"), self)
        self.overrideBox = QCheckBox(self.tr("Override"), self)
        buttonBox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)

        line = 0
        layout.addWidget(self.markColorWidget, line, 0)
        layout.addWidget(self.importCharDrop, line, 3, 1, 2)
        line += 1
        layout.addWidget(self.addGlyphsEdit, line, 0, 1, 5)
        line += 1
        layout.addWidget(self.addUnicodeBox, line, 0)
        layout.addWidget(self.addAsTemplateBox, line, 1)
        layout.addWidget(self.sortFontBox, line, 2)
        layout.addWidget(self.overrideBox, line, 3)
        layout.addWidget(buttonBox, line, 4)
        self.setLayout(layout)

    @classmethod
    def getNewGlyphNames(cls, parent, currentGlyphs=None):
        dialog = cls(currentGlyphs, parent)
        result = dialog.exec_()
        markColor = dialog.markColorWidget.color()
        if markColor is not None:
            markColor = markColor.getRgbF()
        params = dict(
            addUnicode=dialog.addUnicodeBox.isChecked(),
            asTemplate=dialog.addAsTemplateBox.isChecked(),
            markColor=markColor,
            override=dialog.overrideBox.isChecked(),
            sortFont=dialog.sortFontBox.isChecked(),
        )
        newGlyphNames = []
        for name in dialog.addGlyphsEdit.toPlainText().split():
            if name not in dialog.currentGlyphNames:
                newGlyphNames.append(name)
        return (newGlyphNames, params, result)

    def importGlyphs(self, index):
        if index == 0:
            return
        glyphNames = self.importCharDrop.currentData()
        editorNames = self.addGlyphsEdit.toPlainText().split()
        currentNames = set(self.currentGlyphNames) ^ set(editorNames)
        changed = False
        for name in glyphNames:
            if name not in currentNames:
                changed = True
                editorNames.append(name)
        if changed:
            self.addGlyphsEdit.setPlainText(" ".join(editorNames))
            cursor = self.addGlyphsEdit.textCursor()
            cursor.movePosition(QTextCursor.End, QTextCursor.MoveAnchor)
            self.addGlyphsEdit.setTextCursor(cursor)
        self.importCharDrop.setCurrentIndex(0)
        self.addGlyphsEdit.setFocus(Qt.OtherFocusReason)


class SortDialog(QDialog):
    def __init__(self, desc=None, parent=None):
        super().__init__(parent)
        self.setWindowModality(Qt.WindowModal)
        self.setWindowTitle(self.tr("Sort…"))

        self.smartSortBox = QRadioButton(self.tr("Canned sort"), self)
        self.smartSortBox.setToolTip(
            self.tr(
                "A combination of simple, complex and custom "
                "sorts that give optimized ordering results."
            )
        )
        self.glyphSetBox = QRadioButton(self.tr("Glyph set"), self)
        self.glyphSetBox.toggled.connect(self.glyphSetToggle)
        self.glyphSetDrop = QComboBox(self)
        self.glyphSetDrop.setEnabled(False)
        glyphSets = settings.readGlyphSets()
        for name, glyphNames in glyphSets.items():
            self.glyphSetDrop.addItem(name, glyphNames)
        self.customSortBox = QRadioButton(self.tr("Custom…"), self)
        self.customSortBox.toggled.connect(self.customSortToggle)

        self.customSortGroup = QGroupBox(parent=self)
        self.customSortGroup.setEnabled(False)
        descriptorsCount = 6
        if desc is None:
            pass
        elif desc[0]["type"] == "glyphSet":
            self.glyphSetBox.setChecked(True)
            self.glyphSetDrop.setEnabled(True)
            # XXX: we must handle unknown glyphSets... or glyphSets with
            # the same name but different
            self.glyphSetDrop.setCurrentText(desc[0]["name"])
        elif desc[0]["type"] == "cannedDesign":
            self.smartSortBox.setChecked(True)
        else:
            self.customSortBox.setChecked(True)
            self.customSortGroup.setEnabled(True)
            descriptorsCount = len(desc)
        self.customDescriptors = [[] for i in range(descriptorsCount)]
        self.customSortLayout = QGridLayout()
        for i, line in enumerate(self.customDescriptors):
            line.append(QComboBox(self))
            line[0].insertItems(0, sortItems)
            line.append(QCheckBox(self.tr("Ascending"), self))
            line.append(QCheckBox(self.tr("Allow pseudo-unicode"), self))
            if self.customSortBox.isChecked():
                line[0].setCurrentIndex(self.indexFromItemName(desc[i]["type"]))
                line[1].setChecked(desc[i]["ascending"])
                line[2].setChecked(desc[i]["allowPseudoUnicode"])
            else:
                line[0].setCurrentIndex(i)
                line[1].setChecked(True)
                line[2].setChecked(True)
            self.customSortLayout.addWidget(line[0], i, 0)
            self.customSortLayout.addWidget(line[1], i, 1)
            self.customSortLayout.addWidget(line[2], i, 2)
            btn = QPushButton(self)
            btn.setFixedWidth(32)
            btn.setProperty("index", i)
            line.append(btn)
            self.customSortLayout.addWidget(btn, i, 3)
            if i == 0:
                btn.setText("+")
                btn.clicked.connect(self._addRow)
                self.addLineButton = btn
            else:
                btn.setText("−")
                btn.clicked.connect(self._deleteRow)
        self.customSortGroup.setLayout(self.customSortLayout)

        buttonBox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)

        layout = QVBoxLayout(self)
        layout.addWidget(self.smartSortBox)
        layout.addWidget(self.glyphSetBox)
        layout.addWidget(self.glyphSetDrop)
        layout.addWidget(self.customSortBox)
        layout.addWidget(self.customSortGroup)
        layout.addWidget(buttonBox)
        self.setLayout(layout)

    def _addRow(self):
        i = len(self.customDescriptors)
        line = []
        line.append(QComboBox(self))
        line[0].insertItems(0, sortItems)
        line[0].setCurrentIndex(0)
        line.append(QCheckBox("Ascending", self))
        line.append(QCheckBox("Allow pseudo-unicode", self))
        btn = QPushButton("−", self)
        btn.setFixedWidth(32)
        btn.setProperty("index", i)
        btn.clicked.connect(self._deleteRow)
        line.append(btn)
        self.customDescriptors.append(line)
        self.customSortLayout.addWidget(line[0], i, 0)
        self.customSortLayout.addWidget(line[1], i, 1)
        self.customSortLayout.addWidget(line[2], i, 2)
        self.customSortLayout.addWidget(line[3], i, 3)
        if i == 7:
            self.sender().setEnabled(False)

    def _deleteRow(self):
        rel = self.sender().property("index")
        desc = self.customDescriptors
        for i in range(rel + 1, len(desc) - 1):
            desc[i][0].setCurrentIndex(desc[i + 1][0].currentIndex())
            desc[i][1].setChecked(desc[i + 1][1].isChecked())
            desc[i][2].setChecked(desc[i + 1][2].isChecked())
        for elem in desc[-1]:
            elem.setParent(None)
        del self.customDescriptors[-1]
        self.addLineButton.setEnabled(True)
        self.adjustSize()

    def indexFromItemName(self, name):
        for index, item in enumerate(sortItems):
            if name == item:
                return index
        print(self.tr("Unknown descriptor name: %s"), name)
        return 0

    @classmethod
    def getDescriptor(cls, parent, sortDescriptor=None):
        dialog = cls(sortDescriptor, parent)
        result = dialog.exec_()
        if dialog.glyphSetBox.isChecked():
            data = dialog.glyphSetDrop.currentData()
            name = dialog.glyphSetDrop.currentText()
            ret = [dict(type="glyphSet", glyphs=data, name=name)]
        elif dialog.customSortBox.isChecked():
            descriptors = []
            for line in dialog.customDescriptors:
                descriptors.append(
                    dict(
                        type=line[0].currentText(),
                        ascending=line[1].isChecked(),
                        allowPseudoUnicode=line[2].isChecked(),
                    )
                )
            ret = descriptors
        else:
            ret = [dict(type="cannedDesign", allowPseudoUnicode=True)]
        return (ret, result)

    def glyphSetToggle(self):
        checkBox = self.sender()
        self.glyphSetDrop.setEnabled(checkBox.isChecked())

    def customSortToggle(self):
        checkBox = self.sender()
        self.customSortGroup.setEnabled(checkBox.isChecked())



import os
import re
import runpy
import shutil
import stat
import traceback
from collections.abc import Sequence

import fontTools.misc.plistlib
from fontTools.ufoLib import UFOLibError, _UFOBaseIO
from PyQt5.QtGui import QImageReader, QPixmap
from PyQt5.QtWidgets import QApplication

LIB_PATH = "lib"
INFO_FILENAME = "info.plist"
RESOURCES_PATH = "resources"

_infoProperties = {
    "name": "The name of the extension.",
    "developer": "The name of the extension developer.",
    "developerURL": "The URL of the extension developer.",
    "addToMenu": "a dict of *path*, *name*, *shortcut*",
    "launchAtStartup": "Whether *mainScript* should be launched at startup.",
    "mainScript": "The extension entry-point Python file.",
    "tfVersionMajor": "The minimum required TruFont major version number.",
    "tfVersionMinor": "The minimum required TruFont minor version number.",
    "tfVersionPatch": "The minimum required TruFont patch version number.",
    "versionMajor": "The extension major version number.",
    "versionMinor": "The extension minor version number.",
    "versionPatch": "The extension major version number.",
}

_privateAttrsRe = re.compile("^[A-Za-z]{2,6}((?!-)\\.[A-Za-z0-9-]{1,63}(?<!-))+$")


def init_info_property(cls, name, doc):
    setterName = f"_set_{name}"
    getterName = f"_get_{name}"

    def getter(self):
        return self._info.get(name, None)

    getter.__name__ = getterName

    def setter(self, value):
        self._info[name] = value
        if value is None:
            del self._info[name]

    setter.__name__ = setterName

    prop = property(getter, setter, doc)

    setattr(cls, setterName, setter)
    setattr(cls, getterName, getter)
    setattr(cls, name, prop)


def init_info_properties(cls):
    for name, setup in _infoProperties.items():
        init_info_property(cls, name, setup)
    return cls


def remove_readonly(func, path, _):
    "Clear the readonly bit and reattempt the removal"
    os.chmod(path, stat.S_IWRITE)
    func(path)


@init_info_properties
class TExtension:
    def __init__(self, path=None):
        self._info = TExtensionInfo()
        self._path = path
        self._libPath = None
        self._resourcesPath = None

        if path:
            loader = TExtensionReader(self._path)
            loader.readInfo(self._info)

    def __repr__(self):
        path = self._path
        if path is not None and path.rstrip("\\/ ").endswith(".tfExt"):
            fileName = os.path.basename(path)[1]
            return "<%s>" % fileName
        return super().__repr__()

    # info and special-cased getters

    @property
    def info(self):
        return self._info

    def _get_tfVersion(self):
        if self.tfVersionMajor is None:
            return None
        return Version((self.tfVersionMajor, self.tfVersionMinor, self.tfVersionPatch))

    def _set_tfVersion(self, value):
        if value is not None:
            value = Version(value)
        else:
            value = (None, None, None)
        self.tfVersionMajor, self.tfVersionMinor, self.tfVersionPatch = value

    tfVersion = property(
        _get_tfVersion,
        _set_tfVersion,
        doc="The minimum required TruFont version number.",
    )

    def _get_version(self):
        if self.versionMajor is None:
            return None
        return Version((self.versionMajor, self.versionMinor, self.versionPatch))

    def _set_version(self, value):
        if value is not None:
            value = Version(value)
        else:
            value = (None, None, None)
        self.versionMajor, self.versionMinor, self.versionPatch = value

    version = property(_get_version, _set_version, doc="The extension version number.")

    # settings and methods

    @property
    def path(self):
        return self._path

    @property
    def libPath(self):
        return self._libPath

    @libPath.setter
    def libPath(self, path):
        if path is not None:
            assert os.path.isdir(path)
        self._libPath = path

    @property
    def resourcesPath(self):
        return self._resourcesPath

    @resourcesPath.setter
    def resourcesPath(self, path):
        if path is not None:
            assert os.path.isdir(path)
        self._resourcesPath = path

    def save(self, path=None):
        # fetch path
        if path is None or path == self._path:
            return
        # write
        writer = TExtensionWriter(path)
        libPath = self._libPath
        if libPath is None and self._path is not None:
            libPath = os.path.join(self._path, LIB_PATH)
        writer.writeLib(libPath)
        writer.writeInfo(self._info)
        writer.writeResources(os.path.join(path, self._resourcesPath or RESOURCES_PATH))
        # done
        self._path = path

    def get(self, name, ext="*"):
        for file in os.listdir(self._path):
            if ext == "*" or file.endswith(ext):
                imageReader = QImageReader(file)
                if imageReader:
                    return QPixmap.fromImageReader(imageReader)
                return os.path.join(self._path, file)
        return None

    def install(self):
        app = QApplication.instance()
        if self._path is not None:
            fileName = os.path.basename(self._path)
        else:
            fileName = self.name or "Extension" + ".tfExt"
        folder = os.path.join(app.getExtensionsDirectory(), fileName)
        self.save(folder)
        app.registerExtension(self)

    def uninstall(self):
        app = QApplication.instance()
        folder = app.getScriptsDirectory()
        path = self._path
        if path.contains(folder):
            shutil.rmtree(path, onerror=remove_readonly)
            app.unregisterExtension(self)

    def validate(self):
        # TODO: should it e.g. look if the mainScript is an actual Python file
        raise NotImplementedError

    def run(self, subPath=None):
        libPath = self._libPath
        if libPath is None:
            libPath = LIB_PATH
        if subPath is None:
            subPath = self.mainScript
            if subPath is None:
                subPath = ""
        runPath = os.path.join(self._path, libPath, subPath)
        app = QApplication.instance()
        global_vars = app.globals()
        try:
            runpy.run_path(runPath, global_vars, self.name)
        except Exception:
            traceback.print_exc()


class TExtensionInfo(dict):
    def __setitem__(self, key, value):
        if key not in _infoProperties.keys():
            if not _privateAttrsRe.match(key):
                raise AttributeError(
                    "Custom keys can only be added as reverse-domain names."
                )
        super().__setitem__(key, value)


class TExtensionReader:
    def __init__(self, path):
        if not os.path.exists(path):
            raise UFOLibError("The specified extension doesn't exist.")
        self._path = path

    _getPlist = _UFOBaseIO._getPlist

    def _readInfo(self):
        data = self._getPlist(INFO_FILENAME, {})
        if not isinstance(data, dict):
            raise UFOLibError("info.plist is not properly formatted.")
        return data

    def readInfo(self, info):
        infoDict = self._readInfo()
        infoDataToSet = {}
        # standard attrs
        for attr in _infoProperties.keys():
            if attr in infoDict:
                value = infoDict[attr]
                del infoDict[attr]
            else:
                value = None
            if value is None:
                continue
            infoDataToSet[attr] = value
        # remaining private attrs
        for attr in infoDict:
            if not _privateAttrsRe.match(attr):
                continue
            value = infoDict[attr]
            if value is None:
                continue
            infoDataToSet[attr] = value
        for attr, value in infoDataToSet.items():
            try:
                info[attr] = value
            except AttributeError:
                raise UFOLibError(
                    "The supplied info object does not support setting a "
                    "necessary attribute (%s)." % attr
                )


class TExtensionWriter:
    def __init__(self, path):
        self._path = path
        self._makeDirectory()

    def _writePlist(self, fileName, data):
        """
        Write a property list. The errors that
        could be raised during the writing of
        a plist are unpredictable and/or too
        large to list, so, a blind try: except Exception:
        is done. If an exception occurs, a
        UFOLibError will be raised.
        """
        self._makeDirectory()
        path = os.path.join(self._path, fileName)
        try:
            data = fontTools.misc.plistlib.dump(data, path)
        except Exception:
            raise UFOLibError(
                "The data for the file %s could not be written because it is "
                "not properly formatted." % fileName
            )

    def _makeDirectory(self, subDirectory=None):
        path = self._path
        if subDirectory:
            path = os.path.join(self._path, subDirectory)
        if not os.path.exists(path):
            os.makedirs(path)
        return path

    def writeLib(self, origPath):
        if origPath is None:
            return
        canonicalPath = os.path.join(self._path, LIB_PATH)
        # cleanup existing script dir
        if os.path.exists(canonicalPath):
            try:
                shutil.rmtree(canonicalPath, onerror=remove_readonly)
            except Exception:
                raise UFOLibError("Couldn't delete existing script folder.")
        # move in
        try:
            shutil.copytree(origPath, canonicalPath)
        except Exception:
            raise UFOLibError("Couldn't copy script files to the script folder.")

    def writeInfo(self, info):
        infoData = {}
        for key in info:
            if key in _infoProperties or _privateAttrsRe.match(key):
                value = info.get(key, None)
                if value is None:
                    continue
                infoData[key] = value
        # TODO: validate data here?
        # write file
        self._writePlist(INFO_FILENAME, infoData)

    def writeResources(self, origPath):
        pass


class Version(str):
    def __new__(self, value):
        # convert from sequence
        if isinstance(value, Sequence) and not isinstance(value, str):
            assert len(value) == 3
            value = ".".join(str(num) for num in value)
        return super().__new__(Version, value)

    def __iter__(self):
        return iter(_stringToSequence(self))

    def _get_major(self):
        return _stringToSequence(self)[0]

    major = property(_get_major, "The major component.")

    def _get_minor(self):
        return _stringToSequence(self)[1]

    minor = property(_get_minor, "The minor component.")

    def _get_patch(self):
        return _stringToSequence(self)[2]

    patch = property(_get_patch, "The patch component.")


def _stringToSequence(value):
    major, minor, patch = [i.strip() for i in value.split(".")]
    value = []
    for component in (major, minor, patch):
        v = int(component)
        value.append(v)
    return value
